# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Callable, Generator, Sequence
import io
import os
import queue
import time
from typing import Any, ClassVar, Literal, overload
import warnings

import requests

from openstack._utils import renamed_param
from openstack import exceptions
from openstack.image.v2 import cache as _cache
from openstack.image.v2 import image as _image
from openstack.image.v2 import image_location as _image_location
from openstack.image.v2 import image_tasks as _image_tasks
from openstack.image.v2 import member as _member
from openstack.image.v2 import metadef_namespace as _metadef_namespace
from openstack.image.v2 import metadef_object as _metadef_object
from openstack.image.v2 import metadef_property as _metadef_property
from openstack.image.v2 import metadef_resource_type as _metadef_resource_type
from openstack.image.v2 import metadef_schema as _metadef_schema
from openstack.image.v2 import schema as _schema
from openstack.image.v2 import service_info as _si
from openstack.image.v2 import task as _task
from openstack import proxy
from openstack.proxy import CleanupDependency
from openstack import resource
from openstack import utils
from openstack import warnings as os_warnings

# Rackspace returns this for intermittent import errors
_IMAGE_ERROR_396 = "Image cannot be imported. Error code: '396'"
_INT_PROPERTIES = ('min_disk', 'min_ram', 'size', 'virtual_size')
_RAW_PROPERTIES = ('is_protected', 'protected', 'tags')


def _get_name_and_filename(
    name: str, image_format: str
) -> tuple[str, str | None]:
    # See if name points to an existing file
    if os.path.exists(name) and os.path.isfile(name):
        # Neat. Easy enough
        return os.path.splitext(os.path.basename(name))[0], name

    # Try appending the disk format
    name_with_ext = '.'.join((name, image_format))
    if os.path.exists(name_with_ext) and os.path.isfile(name):
        return os.path.basename(name), name_with_ext

    return name, None


class Proxy(proxy.Proxy):
    api_version: ClassVar[Literal['2']] = '2'

    _resource_registry = {
        "cache": _cache.Cache,
        "image": _image.Image,
        "image_member": _member.Member,
        "metadef_namespace": _metadef_namespace.MetadefNamespace,
        "metadef_resource_type": _metadef_resource_type.MetadefResourceType,
        "metadef_resource_type_association": _metadef_resource_type.MetadefResourceTypeAssociation,  # noqa
        "schema": _schema.Schema,
        "info_import": _si.Import,
        "info_store": _si.Store,
        "task": _task.Task,
    }

    retriable_status_codes = [503]

    _IMAGE_MD5_KEY = 'owner_specified.openstack.md5'
    _IMAGE_SHA256_KEY = 'owner_specified.openstack.sha256'
    _IMAGE_OBJECT_KEY = 'owner_specified.openstack.object'

    # NOTE(shade) shade keys were owner_specified.shade.md5 - we need to add
    #             those to freshness checks so that a shade->sdk transition
    #             doesn't result in a re-upload
    _SHADE_IMAGE_MD5_KEY = 'owner_specified.shade.md5'
    _SHADE_IMAGE_SHA256_KEY = 'owner_specified.shade.sha256'
    _SHADE_IMAGE_OBJECT_KEY = 'owner_specified.shade.object'

    # ====== CACHE MANAGEMENT======
    def get_image_cache(self) -> _cache.Cache:
        return self._get(_cache.Cache, requires_id=False)

    def cache_delete_image(
        self,
        image: str | _image.Image,
        ignore_missing: bool = True,
    ) -> None:
        """Delete an image from cache.

        :param image: The value can be either the ID of an image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param ignore_missing: When set to ``False``,
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the image or cache entry does not exist.
        :returns: ``None``
        """
        self._delete(_cache.Cache, image, ignore_missing=ignore_missing)

    @renamed_param('image_id', 'image')
    def queue_image(self, image: str | _image.Image) -> None:
        """Queue image(s) for caching."""
        cache = self._get_resource(_cache.Cache, None)
        cache.queue(self, resource.Resource._get_id(image))

    def clear_cache(self, target: str = 'both') -> None:
        """Clear all images from cache, queue or both

        :param target: Specify which target you want to clear
            One of: ``both``(default), ``cache``, ``queue``.
        """
        cache = self._get_resource(_cache.Cache, None)
        cache.clear(self, target)

    # ====== IMAGES ======

    def _make_v2_image_params(
        self,
        meta: dict[str, Any],
        properties: dict[str, Any],
    ) -> dict[str, Any]:
        ret: dict[str, Any] = {}
        for k, v in iter(properties.items()):
            if k in _INT_PROPERTIES:
                ret[k] = int(v)
            elif k in _RAW_PROPERTIES:
                ret[k] = v
            else:
                if v is None:
                    ret[k] = None
                else:
                    ret[k] = str(v)
        ret.update(meta)
        return ret

    def create_image(
        self,
        name: str,
        *,
        filename: str | None = None,
        data: bytes | None = None,
        container: str | None = None,
        md5: str | None = None,
        sha256: str | None = None,
        disk_format: str | None = None,
        container_format: str | None = None,
        tags: list[str] | None = None,
        disable_vendor_agent: bool = True,
        allow_duplicates: bool = False,
        meta: dict[str, Any] | None = None,
        wait: bool = False,
        timeout: int = 3600,
        validate_checksum: bool = False,
        use_import: bool = False,
        import_method: str | None = None,
        uri: str | None = None,
        remote_region: str | None = None,
        remote_image_id: str | None = None,
        remote_service_interface: str | None = None,
        stores: list[str] | None = None,
        all_stores: bool | None = None,
        all_stores_must_succeed: bool | None = None,
        **kwargs: Any,
    ) -> _image.Image:
        """Create an image and optionally upload data

        Create a new image. If ``filename`` or ``data`` are provided, it will
        also upload data to this image.

        Note that uploading image data is actually quite a complicated
        procedure. There are three ways to upload an image:

        * Image upload
        * Image import
        * Image tasks

        If the image tasks API is enabled, this must be used. However, this API
        is deprecated since the Image service's Mitaka (12.0.0) release and is
        now admin-only. Assuming this API is not enabled, you may choose
        between image upload or image import. Image import is more powerful and
        allows you to upload data from multiple sources including other glance
        instances. It should be preferred on all services that support it.

        :param name: Name of the image to create. If it is a pathname
            of an image, the name will be constructed from the extensionless
            basename of the path.
        :param filename: The path to the file to upload, if needed.
            (optional, defaults to None)
        :param data: Image data (string or file-like object). It is mutually
            exclusive with filename
        :param container: Name of the container in swift where images
            should be uploaded for import if the cloud requires such a thing.
            (optional, defaults to 'images')
        :param md5: md5 sum of the image file. If not given, an md5 will
            be calculated.
        :param sha256: sha256 sum of the image file. If not given, an md5
            will be calculated.
        :param disk_format: The disk format the image is in. (optional,
            defaults to the os-client-config config value for this cloud)
        :param container_format: The container format the image is in.
            (optional, defaults to the os-client-config config value for this
            cloud)
        :param tags: List of tags for this image. Each tag is a string
            of at most 255 chars.
        :param disable_vendor_agent: Whether or not to append metadata
            flags to the image to inform the cloud in question to not expect a
            vendor agent to be runing. (optional, defaults to True)
        :param allow_duplicates: If true, skips checks that enforce unique
            image name. (optional, defaults to False)
        :param meta: A dict of key/value pairs to use for metadata that
            bypasses automatic type conversion.
        :param wait: If true, waits for image to be created. Defaults to
            true - however, be aware that one of the upload methods is always
            synchronous.
        :param timeout: Seconds to wait for image creation. None is forever.
        :param validate_checksum: If true and cloud returns checksum,
            compares return value with the one calculated or passed into this
            call. If value does not match - raises exception. Default is
            'false'
        :param use_import: Use the 'glance-direct' method of the
            interoperable image import mechanism to import the image. This
            defaults to false because it is harder on the target cloud so
            should only be used when needed, such as when the user needs the
            cloud to transform image format. If the cloud has disabled direct
            uploads, this will default to true. If you wish to use other import
            methods, use the ``import_image`` method instead.
        :param import_method: Method to use for importing the image. Not all
            deployments support all methods. One of: ``glance-direct``
            (default), ``web-download``, ``glance-download`` (``copy-image`` is
            not used with create). Use of ``glance-direct`` requires the image
            be first staged.
        :param uri: Required only if using the ``web-download`` import method.
            This url is where the data is made available to the Image
            service.
        :param remote_region: The remote glance region to download the image
            from when using glance-download.
        :param remote_image_id: The ID of the image to import from the
            remote glance when using glance-download.
        :param remote_service_interface: The remote glance service interface to
            use when using glance-download.
        :param stores: List of stores to be used when enabled_backends is
            activated in glance. List values can be the id of a store or a
            :class:`~openstack.image.v2.service_info.Store` instance.
            Implies ``use_import`` equals ``True``.
        :param all_stores: Upload to all available stores. Mutually exclusive
            with ``store`` and ``stores``.
            Implies ``use_import`` equals ``True``.
        :param all_stores_must_succeed: When set to True, if an error occurs
            during the upload in at least one store, the worfklow fails, the
            data is deleted from stores where copying is done (not staging),
            and the state of the image is unchanged. When set to False, the
            workflow will fail (data deleted from stores, …) only if the import
            fails on all stores specified by the user. In case of a partial
            success, the locations added to the image will be the stores where
            the data has been correctly uploaded.
            Default is True.
            Implies ``use_import`` equals ``True``.

        Additional kwargs will be passed to the image creation as additional
        metadata for the image and will have all values converted to string
        except for min_disk, min_ram, size and virtual_size which will be
        converted to int.

        If you are sure you have all of your data types correct or have an
        advanced need to be explicit, use meta. If you are just a normal
        consumer, using kwargs is likely the right choice.

        If a value is in meta and kwargs, meta wins.

        :returns: The results of image creation
        :raises: SDKException if there are problems uploading
        """
        if filename and data:
            raise exceptions.SDKException(
                'filename and data are mutually exclusive'
            )

        if container is None:
            container = self._connection._OBJECT_AUTOCREATE_CONTAINER

        if not meta:
            meta = {}

        if not disk_format:
            disk_format = self._connection.config.config['image_format']

        if not container_format:
            # https://docs.openstack.org/image-guide/image-formats.html
            container_format = 'bare'

        # If there is no filename, see if name is actually the filename
        if not filename and not data:
            name, filename = _get_name_and_filename(
                name,
                self._connection.config.config['image_format'],
            )

        if validate_checksum and data and not isinstance(data, bytes):
            raise exceptions.SDKException(
                'Validating checksum is not possible when data is not a '
                'direct binary object'
            )

        if not (md5 or sha256) and validate_checksum:
            if filename:
                md5, sha256 = utils._get_file_hashes(filename)
            elif data and isinstance(data, bytes):
                md5, sha256 = utils._calculate_data_hashes(data)

        if allow_duplicates:
            current_image = None
        else:
            current_image = self.find_image(name)
            if current_image:
                # NOTE(pas-ha) 'properties' may be absent or be None
                props = current_image.get('properties') or {}
                md5_key = props.get(
                    self._IMAGE_MD5_KEY,
                    props.get(self._SHADE_IMAGE_MD5_KEY, ''),
                )
                sha256_key = props.get(
                    self._IMAGE_SHA256_KEY,
                    props.get(self._SHADE_IMAGE_SHA256_KEY, ''),
                )
                up_to_date = utils._hashes_up_to_date(
                    md5=md5,
                    sha256=sha256,
                    md5_key=md5_key,
                    sha256_key=sha256_key,
                )
                if up_to_date:
                    self.log.debug(
                        "image %(name)s exists and is up to date",
                        {'name': name},
                    )
                    return current_image
                else:
                    self.log.debug(
                        "image %(name)s exists, but contains different "
                        "checksums. Updating.",
                        {'name': name},
                    )

        if disable_vendor_agent:
            kwargs.update(
                self._connection.config.config['disable_vendor_agent']
            )

        # If a user used the v1 calling format, they will have
        # passed a dict called properties along
        properties = kwargs.pop('properties', {})
        properties[self._IMAGE_MD5_KEY] = md5 or ''
        properties[self._IMAGE_SHA256_KEY] = sha256 or ''
        properties[self._IMAGE_OBJECT_KEY] = '/'.join([container, name])
        kwargs.update(properties)
        image_kwargs: dict[str, Any] = {'properties': kwargs}
        if disk_format:
            image_kwargs['disk_format'] = disk_format
        if container_format:
            image_kwargs['container_format'] = container_format
        if tags:
            image_kwargs['tags'] = tags

        if filename or data or import_method:
            image = self._upload_image(
                name,
                filename=filename,
                data=data,
                meta=meta,
                wait=wait,
                timeout=timeout,
                validate_checksum=validate_checksum,
                use_import=use_import,
                import_method=import_method,
                uri=uri,
                remote_region=remote_region,
                remote_image_id=remote_image_id,
                remote_service_interface=remote_service_interface,
                stores=stores,
                all_stores=all_stores,
                all_stores_must_succeed=all_stores_must_succeed,
                **image_kwargs,
            )
        else:
            properties = image_kwargs.pop('properties', {})
            image_kwargs.update(self._make_v2_image_params(meta, properties))
            image_kwargs['name'] = name
            image = self._create(_image.Image, **image_kwargs)

        return image

    def import_image(
        self,
        image: str | _image.Image,
        method: str = 'glance-direct',
        *,
        uri: str | None = None,
        remote_region: str | None = None,
        remote_image_id: str | None = None,
        remote_service_interface: str | None = None,
        store: str | _si.Store | None = None,
        stores: list[str | _si.Store] | None = None,
        all_stores: bool | None = None,
        all_stores_must_succeed: bool | None = None,
    ) -> requests.Response:
        """Import data to an existing image

        Interoperable image import process are introduced in the Image API
        v2.6. It mainly allow image importing from an external url and let
        Image Service download it by itself without sending binary data at
        image creation.

        :param image: The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param method: Method to use for importing the image. Not all
            deployments support all methods. One of: ``glance-direct``
            (default), ``web-download``, ``glance-download``, or
            ``copy-image``. Use of ``glance-direct`` requires the image be
            first staged.
        :param uri: Required only if using the ``web-download`` import method.
            This url is where the data is made available to the Image
            service.
        :param remote_region: The remote glance region to download the image
            from when using glance-download.
        :param remote_image_id: The ID of the image to import from the
            remote glance when using glance-download.
        :param remote_service_interface: The remote glance service interface to
            use when using glance-download.
        :param store: Used when enabled_backends is activated in glance. The
            value can be the id of a store or a.
            :class:`~openstack.image.v2.service_info.Store` instance.
        :param stores: List of stores to be used when enabled_backends is
            activated in glance. List values can be the id of a store or a
            :class:`~openstack.image.v2.service_info.Store` instance.
        :param all_stores: Upload to all available stores. Mutually exclusive
            with ``store`` and ``stores``.
        :param all_stores_must_succeed: When set to True, if an error occurs
            during the upload in at least one store, the worfklow fails, the
            data is deleted from stores where copying is done (not staging),
            and the state of the image is unchanged. When set to False, the
            workflow will fail (data deleted from stores, …) only if the
            import fails on all stores specified by the user. In case of
            a partial success, the locations added to the image will be
            the stores where the data has been correctly uploaded.
            Default is True.

        :returns: The raw response from the request.
        """
        image = self._get_resource(_image.Image, image)

        if all_stores and (store or stores):
            raise exceptions.InvalidRequest(
                "all_stores is mutually exclusive with store and stores"
            )

        if store is not None:
            if stores:
                raise exceptions.InvalidRequest(
                    "store and stores are mutually exclusive"
                )
            store = self._get_resource(_si.Store, store)

        stores = stores or []
        new_stores: list[str | _si.Store] = []
        for s in stores:
            new_stores.append(self._get_resource(_si.Store, s))
        stores = new_stores

        # as for the standard image upload function, container_format and
        # disk_format are required for using image import process
        if not all([image.container_format, image.disk_format]):
            raise exceptions.InvalidRequest(
                "Both container_format and disk_format are required for "
                "importing an image"
            )

        return image.import_image(
            self,
            method=method,
            uri=uri,
            remote_region=remote_region,
            remote_image_id=remote_image_id,
            remote_service_interface=remote_service_interface,
            store=store,
            stores=stores,
            all_stores=all_stores,
            all_stores_must_succeed=all_stores_must_succeed,
        )

    def stage_image(
        self,
        image: str | _image.Image,
        *,
        filename: str | None = None,
        data: Any = None,
    ) -> _image.Image:
        """Stage binary image data

        :param image: The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param filename: Optional name of the file to read data from.
        :param data: Optional data to be uploaded as an image.

        :returns: The results of image creation
        """
        if filename and data:
            raise exceptions.SDKException(
                'filename and data are mutually exclusive'
            )

        image = self._get_resource(_image.Image, image)

        if 'queued' != image.status:
            raise exceptions.SDKException(
                'Image stage is only possible for images in the queued state. '
                f'Current state is {image.status}'
            )

        if filename:
            image.data = open(filename, 'rb')
        elif data:
            image.data = data
        image.stage(self)

        # Stage does not return content, but updates the object
        image.fetch(self)

        return image

    def upload_image(
        self,
        container_format: str | None = None,
        disk_format: str | None = None,
        data: Any = None,
        **attrs: Any,
    ) -> _image.Image:
        """Create and upload a new image from attributes

        .. warning:

           This method is deprecated - and also doesn't work very well.
           Please stop using it immediately and switch to `create_image`.

        :param container_format: Format of the container.
            A valid value is ami, ari, aki, bare, ovf, ova, or docker.
        :param disk_format: The format of the disk. A valid value is ami,
            ari, aki, vhd, vmdk, raw, qcow2, vdi, or iso.
        :param data: The data to be uploaded as an image.
        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.image.v2.image.Image`, comprised of the
            properties on the Image class.

        :returns: The results of image creation
        """
        warnings.warn(
            "upload_image is deprecated. Use create_image instead.",
            os_warnings.RemovedInSDK50Warning,
        )
        # container_format and disk_format are required to be set
        # on the image by the time upload_image is called, but they're not
        # required by the _create call. Enforce them here so that we don't
        # need to handle a failure in _create, as upload_image will
        # return a 400 with a message about disk_format and container_format
        # not being set.
        if not all([container_format, disk_format]):
            raise exceptions.InvalidRequest(
                "Both container_format and disk_format are required"
            )

        img = self._create(
            _image.Image,
            disk_format=disk_format,
            container_format=container_format,
            **attrs,
        )

        # TODO(briancurtin): Perhaps we should run img.upload_image
        # in a background thread and just return what is called by
        # self._create, especially because the upload_image call doesn't
        # return anything anyway. Otherwise this blocks while uploading
        # significant amounts of image data.
        img.data = data
        img.upload(self)

        return img

    def _upload_image(
        self,
        name: str,
        *,
        filename: str | None = None,
        data: Any = None,
        meta: dict[str, Any] | None = None,
        wait: bool = False,
        timeout: int | None = None,
        validate_checksum: bool = True,
        use_import: bool = False,
        import_method: str | None = None,
        uri: str | None = None,
        remote_region: str | None = None,
        remote_image_id: str | None = None,
        remote_service_interface: str | None = None,
        stores: list[Any] | None = None,
        all_stores: bool | None = None,
        all_stores_must_succeed: bool | None = None,
        **kwargs: Any,
    ) -> _image.Image:
        # We can never have nice things. Glance v1 took "is_public" as a
        # boolean. Glance v2 takes "visibility". If the user gives us
        # is_public, we know what they mean. If they give us visibility, they
        # know that they mean.
        if 'is_public' in kwargs['properties']:
            warnings.warn(
                "The 'is_public' property is not supported by Glance v2: use "
                "'visibility=public/private' instead",
                os_warnings.RemovedInSDK60Warning,
            )

            is_public = kwargs['properties'].pop('is_public')
            if is_public:
                kwargs['visibility'] = 'public'
            else:
                kwargs['visibility'] = 'private'

        try:
            # This makes me want to die inside
            if self._connection.image_api_use_tasks:
                if use_import:
                    raise exceptions.SDKException(
                        "The Glance Task API and Import API are mutually "
                        "exclusive. Either disable image_api_use_tasks in "
                        "config, or do not request using import"
                    )
                return self._upload_image_task(  # type: ignore[return-value]
                    name,
                    filename,
                    data=data,
                    meta=meta,
                    wait=wait,
                    timeout=timeout,
                    **kwargs,
                )
            else:
                return self._upload_image_put(
                    name,
                    filename,
                    data=data,
                    meta=meta,
                    validate_checksum=validate_checksum,
                    use_import=use_import,
                    import_method=import_method,
                    uri=uri,
                    remote_region=remote_region,
                    remote_image_id=remote_image_id,
                    remote_service_interface=remote_service_interface,
                    stores=stores,
                    all_stores=all_stores,
                    all_stores_must_succeed=all_stores_must_succeed,
                    **kwargs,
                )
        except exceptions.SDKException:
            self.log.debug("Image creation failed", exc_info=True)
            raise
        except Exception as e:
            raise exceptions.SDKException(
                f"Image creation failed: {e!s}"
            ) from e

    def _upload_image_put(
        self,
        name: str,
        filename: str | None,
        data: Any,
        meta: dict[str, Any] | None,
        validate_checksum: bool,
        use_import: bool = False,
        import_method: str | None = None,
        uri: str | None = None,
        remote_region: str | None = None,
        remote_image_id: str | None = None,
        remote_service_interface: str | None = None,
        stores: list[Any] | None = None,
        all_stores: bool | None = None,
        all_stores_must_succeed: bool | None = None,
        **image_kwargs: Any,
    ) -> _image.Image:
        if all_stores and stores:
            raise exceptions.InvalidRequest(
                "all_stores is mutually exclusive with stores"
            )

        # use of any of these imply use_import=True
        if stores or all_stores or all_stores_must_succeed:
            use_import = True

        if use_import and not import_method:
            import_method = 'glance-direct'

        if filename and not data:
            image_data = open(filename, 'rb')
        else:
            image_data = data

        properties = image_kwargs.pop('properties', {})

        image_kwargs.update(self._make_v2_image_params(meta or {}, properties))
        image_kwargs['name'] = name

        image = self._create(_image.Image, **image_kwargs)
        image.data = image_data

        supports_import = (
            image.image_import_methods
            and import_method in image.image_import_methods
        )
        if use_import and not supports_import:
            raise exceptions.SDKException(
                "Importing image was requested but the cloud does not "
                "support the image import method."
            )

        try:
            if not use_import:
                response = image.upload(self)
                exceptions.raise_from_response(response)
            if use_import:
                kwargs: dict[str, Any] = {}
                if stores is not None:
                    kwargs['stores'] = stores
                else:
                    kwargs['all_stores'] = all_stores
                    kwargs['all_stores_must_succeed'] = all_stores_must_succeed
                if import_method == 'glance-direct':
                    image.stage(self)
                elif import_method == 'web-download':
                    kwargs['uri'] = uri
                elif import_method == 'glance-download':
                    kwargs.update(
                        remote_region=remote_region,
                        remote_image_id=remote_image_id,
                        remote_service_interface=remote_service_interface,
                    )
                assert import_method is not None
                self.import_image(image, method=import_method, **kwargs)

            # image_kwargs are flat here
            md5 = image_kwargs.get(self._IMAGE_MD5_KEY)
            sha256 = image_kwargs.get(self._IMAGE_SHA256_KEY)
            if validate_checksum and (md5 or sha256):
                # Verify that the hash computed remotely matches the local
                # value
                data = image.fetch(self)
                checksum = data.get('checksum')
                if checksum:
                    valid = checksum == md5 or checksum == sha256
                    if not valid:
                        raise Exception('Image checksum verification failed')
        except Exception:
            self.log.debug("Deleting failed upload of image %s", name)
            self.delete_image(image.id)
            raise

        return image

    def _upload_image_task(
        self,
        name: str,
        filename: str | None,
        data: Any,
        wait: bool,
        timeout: int | None,
        meta: dict[str, Any] | None,
        **image_kwargs: Any,
    ) -> _image.Image | _task.Task:
        if not self._connection.has_service('object-store'):
            raise exceptions.SDKException(
                f"The cloud {self._connection.config.name} is configured to "
                f"use tasks for image upload, but no object-store service is "
                f"available. Aborting."
            )

        properties = image_kwargs.get('properties', {})
        md5 = properties[self._IMAGE_MD5_KEY]
        sha256 = properties[self._IMAGE_SHA256_KEY]
        container = properties[self._IMAGE_OBJECT_KEY].split('/', 1)[0]
        image_kwargs.pop('disk_format', None)
        image_kwargs.pop('container_format', None)

        self._connection.create_container(container)  # type: ignore[no-untyped-call]
        self._connection.create_object(  # type: ignore[no-untyped-call]
            container,
            name,
            filename,
            md5=md5,
            sha256=sha256,
            data=data,
            metadata={self._connection._OBJECT_AUTOCREATE_KEY: 'true'},
            **{
                'content-type': 'application/octet-stream',
                'x-delete-after': str(24 * 60 * 60),
            },
        )
        # TODO(mordred): Can we do something similar to what nodepool does
        # using glance properties to not delete then upload but instead make a
        # new "good" image and then mark the old one as "bad"
        task_args = {
            'type': 'import',
            'input': {
                'import_from': f'{container}/{name}',
                'image_properties': {'name': name},
            },
        }

        glance_task = self.create_task(**task_args)
        if wait:
            start = time.time()

            try:
                glance_task = self.wait_for_task(
                    task=glance_task, status='success', wait=timeout
                )

                image_id = glance_task.result['image_id']
                image = self.get_image(image_id)
                # NOTE(gtema): Since we might move unknown attributes of
                # the image under properties - merge current with update
                # properties not to end up removing "existing" properties
                props = image.properties.copy()
                props.update(image_kwargs.pop('properties', {}))
                image_kwargs['properties'] = props

                image = self.update_image(image, **image_kwargs)
                self.log.debug(
                    "Image Task %s imported %s in %s",
                    glance_task.id,
                    image_id,
                    (time.time() - start),
                )
            except exceptions.ResourceFailure as e:
                glance_task = self.get_task(glance_task)
                raise exceptions.SDKException(
                    f"Image creation failed: {e.message}",
                    extra_data=glance_task,
                ) from e
            finally:
                # Clean up after ourselves. The object we created is not
                # needed after the import is done.
                self._connection.delete_object(container, name)  # type: ignore[no-untyped-call]
            return image
        else:
            return glance_task

    def _existing_image(self, **kwargs: Any) -> _image.Image:
        return _image.Image.existing(connection=self._connection, **kwargs)

    def download_image(
        self,
        image: str | _image.Image,
        *,
        stream: bool = False,
        output: str | io.IOBase | None = None,
        chunk_size: int = 1024 * 1024,
        store_preferences: Sequence[str] | None = None,
    ) -> requests.Response:
        """Download an image

        This will download an image to memory when ``stream=False``, or allow
        streaming downloads using an iterator when ``stream=True``.
        For examples of working with streamed responses, see
        :ref:`download_image-stream-true`.

        :param image: The value can be either the ID of an image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param stream: When ``True``, return a :class:`requests.Response`
            instance allowing you to iterate over the response data stream
            instead of storing its entire contents in memory. See
            :meth:`requests.Response.iter_content` for more details.

            *NOTE*: If you do not consume the entirety of the response you must
            explicitly call :meth:`requests.Response.close` or otherwise risk
            inefficiencies with the ``requests`` library's handling of
            connections.

            When ``False``, return the entire contents of the response.
        :param output: Either a file object or a path to store data into.
        :param chunk_size: size in bytes to read from the wire and buffer
            at one time. Defaults to 1024 * 1024 = 1 MiB
        :param store_preferences: List of store names to prefer when
            downloading.

        :returns: When output is not given - the bytes comprising the given
            Image when stream is False, otherwise a :class:`requests.Response`
            instance. When output is given - a
            :class:`~openstack.image.v2.image.Image` instance.
        """

        image = self._get_resource(_image.Image, image)

        return image.download(
            self,
            stream=stream,
            output=output,
            chunk_size=chunk_size,
            store_preferences=store_preferences,
        )

    def delete_image(
        self,
        image: str | _image.Image,
        *,
        store: str | _si.Store | None = None,
        ignore_missing: bool = True,
    ) -> None:
        """Delete an image

        :param image: The value can be either the ID of an image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param store: The value can be either the ID of a store or a
            :class:`~openstack.image.v2.service_info.Store` instance that the
            image is associated with. If specified, the image will only be
            deleted from the specified store.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the image does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent image.

        :returns: ``None``
        """
        if store:
            store = self._get_resource(_si.Store, store)
            store.delete_image(self, image, ignore_missing=ignore_missing)
        else:
            self._delete(_image.Image, image, ignore_missing=ignore_missing)

    @overload
    def find_image(
        self,
        name_or_id: str,
        ignore_missing: Literal[False],
    ) -> _image.Image: ...

    @overload
    def find_image(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _image.Image | None: ...

    def find_image(
        self,
        name_or_id: str,
        ignore_missing: bool = True,
    ) -> _image.Image | None:
        """Find a single image

        :param name_or_id: The name or ID of a image.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.image.v2.image.Image` or None
        """
        return self._find(
            _image.Image,
            name_or_id,
            ignore_missing=ignore_missing,
        )

    def get_image(self, image: str | _image.Image) -> _image.Image:
        """Get a single image

        :param image: The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.

        :returns: One :class:`~openstack.image.v2.image.Image`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_image.Image, image)

    def images(self, **query: Any) -> Generator[_image.Image, None, None]:
        """Return a generator of images

        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of image objects
        """
        return self._list(_image.Image, **query)

    def update_image(
        self, image: str | _image.Image, **attrs: Any
    ) -> _image.Image:
        """Update a image

        :param image: Either the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param attrs: The attributes to update on the image represented
            by ``image``.

        :returns: The updated image
        """
        return self._update(_image.Image, image, **attrs)

    def add_image_location(
        self,
        image: str | _image.Image,
        url: str,
        *,
        validation_data: dict[str, Any] | None = None,
    ) -> _image_location.ImageLocation:
        """Add location to an image

        :param image: Either the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param url: A valid location url
        :param validation_data: Values to be used to populate the
            corresponding image properties.

        :returns: The created image location
        """
        if not url:
            raise exceptions.InvalidRequest(
                "Location url is required to add the location to image."
            )
        if validation_data is None:
            validation_data = {}

        image_id = resource.Resource._get_id(image)

        return self._create(
            _image_location.ImageLocation,
            image_id=image_id,
            url=url,
            validation_data=validation_data,
        )

    def image_locations(
        self,
        image: str | _image.Image,
    ) -> Generator[_image_location.ImageLocation, None, None]:
        """Get locations associated to an Image

        :param image: Either the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :returns: A generator of image location objects
        """
        image_id = resource.Resource._get_id(image)

        return self._list(_image_location.ImageLocation, image_id=image_id)

    def deactivate_image(self, image: str | _image.Image) -> None:
        """Deactivate an image

        :param image: Either the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.

        :returns: None
        """
        image = self._get_resource(_image.Image, image)
        image.deactivate(self)

    def reactivate_image(self, image: str | _image.Image) -> None:
        """Reactivate an image

        :param image: Either the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.

        :returns: None
        """
        image = self._get_resource(_image.Image, image)
        image.reactivate(self)

    def update_image_properties(
        self,
        image: str | _image.Image | None = None,
        meta: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> bool:
        """Update the properties of an existing image

        :param image: The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param meta: A dict of key/value pairs to use for metadata that
            bypasses automatic type conversion.

        Additional kwargs will be passed to the image creation as additional
        metadata for the image and will have all values converted to string
        except for min_disk, min_ram, size and virtual_size which will be
        converted to int.
        """
        image = self._get_resource(_image.Image, image)

        if not meta:
            meta = {}

        properties = {}
        for k, v in iter(kwargs.items()):
            if v and k in ['ramdisk', 'kernel']:
                v = self._connection.get_image_id(v)  # type: ignore[no-untyped-call]
                k = f'{k}_id'
            properties[k] = v

        img_props = image.properties.copy()

        for k, v in iter(self._make_v2_image_params(meta, properties).items()):
            if image.get(k, None) != v:
                img_props[k] = v
        if not img_props:
            return False

        self.update_image(image, **img_props)

        return True

    def image_tasks(
        self,
        image: str | _image.Image,
    ) -> Generator[_image_tasks.ImageTasks, None, None]:
        """Return a generator of Image Tasks

        :param image: The value can be either the name of an image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :returns: A generator object of image tasks
        :raises: :class:`~openstack.exceptions.NotFoundException`
                when no resource can be found.
        """
        image_id = resource.Resource._get_id(image)
        return self._list(_image_tasks.ImageTasks, image_id=image_id)

    def add_tag(self, image: str | _image.Image, tag: str) -> None:
        """Add a tag to an image

        :param image: The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param tag: The tag to be added.

        :returns: None
        """
        image = self._get_resource(_image.Image, image)
        image.add_tag(self, tag)

    def remove_tag(self, image: str | _image.Image, tag: str) -> None:
        """Remove a tag from an image

        :param image: The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param tag: The tag to be removed.

        :returns: None
        """
        image = self._get_resource(_image.Image, image)
        image.remove_tag(self, tag)

    # ====== IMAGE MEMBERS ======
    def add_member(
        self, image: str | _image.Image, **attrs: Any
    ) -> _member.Member:
        """Create a new member from attributes

        :param image: The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance
            that the member will be created for.
        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.image.v2.member.Member`,
            comprised of the properties on the Member class.

        See `Image Sharing Reference
        <https://docs.openstack.org/api-ref/image/v2/index.html?expanded=create-image-member-detail#create-image-member>`__
        for details.

        :returns: The results of member creation
        """
        image_id = resource.Resource._get_id(image)
        return self._create(_member.Member, image_id=image_id, **attrs)

    def remove_member(
        self,
        member: str | _member.Member,
        image: str | _image.Image | None = None,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a member

        :param member: The value can be either the ID of a member or a
            :class:`~openstack.image.v2.member.Member` instance.
        :param image: The value can be either the ID of an image or a
            :class:`~openstack.image.v2.image.Image` instance that the member
            is part of. This is required if ``member`` is an ID.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the member does not exist. When set to ``True``, no exception
            will be set when attempting to delete a nonexistent member.

        :returns: ``None``
        """
        image_id = resource.Resource._get_id(image)  # type: ignore[arg-type]
        member_id = resource.Resource._get_id(member)
        self._delete(
            _member.Member,
            None,
            member_id=member_id,
            image_id=image_id,
            ignore_missing=ignore_missing,
        )

    @overload
    def find_member(
        self,
        name_or_id: str,
        image: str | _image.Image,
        ignore_missing: Literal[False],
    ) -> _member.Member: ...

    @overload
    def find_member(
        self,
        name_or_id: str,
        image: str | _image.Image,
        ignore_missing: bool = True,
    ) -> _member.Member | None: ...

    def find_member(
        self,
        name_or_id: str,
        image: str | _image.Image,
        ignore_missing: bool = True,
    ) -> _member.Member | None:
        """Find a single member

        :param name_or_id: The name or ID of a member.
        :param image: This is the image that the member belongs to,
            the value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.image.v2.member.Member` or None
        """
        image_id = resource.Resource._get_id(image)
        return self._find(
            _member.Member,
            name_or_id,
            image_id=image_id,
            ignore_missing=ignore_missing,
        )

    def get_member(
        self,
        member: str | _member.Member,
        image: str | _image.Image,
    ) -> _member.Member:
        """Get a single member on an image

        :param member: The value can be the ID of a member or a
            :class:`~openstack.image.v2.member.Member` instance.
        :param image: This is the image that the member belongs to.
            The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :returns: One :class:`~openstack.image.v2.member.Member`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        member_id = resource.Resource._get_id(member)
        image_id = resource.Resource._get_id(image)
        return self._get(
            _member.Member, member_id=member_id, image_id=image_id
        )

    def members(
        self,
        image: str | _image.Image,
        **query: Any,
    ) -> Generator[_member.Member, None, None]:
        """Return a generator of members

        :param image: This is the image that the member belongs to,
            the value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of member objects
        """
        image_id = resource.Resource._get_id(image)
        return self._list(_member.Member, image_id=image_id)

    def update_member(
        self,
        member: str | _member.Member,
        image: str | _image.Image,
        **attrs: Any,
    ) -> _member.Member:
        """Update the member of an image

        :param member: Either the ID of a member or a
            :class:`~openstack.image.v2.member.Member` instance.
        :param image: This is the image that the member belongs to.
            The value can be the ID of a image or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param attrs: The attributes to update on the member represented
            by ``member``.

        See `Image Sharing Reference
        <https://docs.openstack.org/api-ref/image/v2/index.html?expanded=update-image-member-detail#update-image-member>`__
        for details.

        :returns: The updated member
        """
        member_id = resource.Resource._get_id(member)
        image_id = resource.Resource._get_id(image)
        return self._update(
            _member.Member,
            None,
            member_id=member_id,
            image_id=image_id,
            **attrs,
        )

    # ====== METADEF NAMESPACES ======
    def create_metadef_namespace(
        self, **attrs: Any
    ) -> _metadef_namespace.MetadefNamespace:
        """Create a new metadef namespace from attributes

        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            comprised of the properties on the MetadefNamespace class.

        :returns: The results of metadef namespace creation
        """
        return self._create(_metadef_namespace.MetadefNamespace, **attrs)

    def delete_metadef_namespace(
        self,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a metadef namespace

        :param metadef_namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :param ignore_missing: When set to ``False``,
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the metadef namespace does not exist.
        :returns: ``None``
        """
        self._delete(
            _metadef_namespace.MetadefNamespace,
            metadef_namespace,
            ignore_missing=ignore_missing,
        )

    # NOTE(stephenfin): There is no 'find_metadef_namespace' since namespaces
    # are identified by the namespace name, not an arbitrary UUID, meaning
    # 'find_metadef_namespace' would be identical to 'get_metadef_namespace'

    def get_metadef_namespace(
        self,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
    ) -> _metadef_namespace.MetadefNamespace:
        """Get a single metadef namespace

        :param metadef_namespace: Either the name of a metadef namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.

        :returns: One
            :class:`~~openstack.image.v2.metadef_namespace.MetadefNamespace`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        return self._get(
            _metadef_namespace.MetadefNamespace,
            metadef_namespace,
        )

    def metadef_namespaces(
        self,
        **query: Any,
    ) -> Generator[_metadef_namespace.MetadefNamespace, None, None]:
        """Return a generator of metadef namespaces

        :returns: A generator object of metadef namespaces
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._list(_metadef_namespace.MetadefNamespace, **query)

    def update_metadef_namespace(
        self,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        **attrs: Any,
    ) -> _metadef_namespace.MetadefNamespace:
        """Update a server

        :param metadef_namespace: Either the name of a metadef namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :param attrs: The attributes to update on the metadef namespace
            represented by ``metadef_namespace``.

        :returns: The updated metadef namespace
        """
        # rather annoyingly, Glance insists on us providing the 'namespace'
        # argument, even if we're not changing it...
        if 'namespace' not in attrs:
            attrs['namespace'] = resource.Resource._get_id(metadef_namespace)

        return self._update(
            _metadef_namespace.MetadefNamespace,
            metadef_namespace,
            **attrs,
        )

    def add_tag_to_metadef_namespace(
        self,
        namespace: str | _metadef_namespace.MetadefNamespace,
        tag: str,
    ) -> None:
        """Add a tag to a metadef namespace

        :param metadef_namespace: Either the name of a metadef namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :param tag: The tag to be added.

        :returns: None
        """
        namespace = self._get_resource(
            _metadef_namespace.MetadefNamespace, namespace
        )
        namespace.add_tag(self, tag)

    def remove_tag_from_metadef_namespace(
        self,
        namespace: str | _metadef_namespace.MetadefNamespace,
        tag: str,
    ) -> None:
        """Remove a tag from a metadef namespace

        :param metadef_namespace: Either the name of a metadef namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :param tag: The tag to be removed.

        :returns: None
        """
        namespace = self._get_resource(
            _metadef_namespace.MetadefNamespace, namespace
        )
        namespace.remove_tag(self, tag)

    def remove_tags_from_metadef_namespace(
        self, namespace: str | _metadef_namespace.MetadefNamespace
    ) -> None:
        """Remove all tags from a metadef namespace

        :param metadef_namespace: Either the name of a metadef namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.

        :returns: None
        """
        namespace = self._get_resource(
            _metadef_namespace.MetadefNamespace, namespace
        )
        namespace.remove_all_tags(self)

    # ====== METADEF OBJECT ======
    def create_metadef_object(
        self,
        namespace: str | _metadef_namespace.MetadefNamespace,
        **attrs: Any,
    ) -> _metadef_object.MetadefObject:
        """Create a new object from namespace

        :param namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.image.v2.metadef_object.MetadefObject`,
            comprised of the properties on the Metadef object class.

        :returns: A metadef namespace
        """
        namespace_name = resource.Resource._get_id(namespace)
        return self._create(
            _metadef_object.MetadefObject,
            namespace_name=namespace_name,
            **attrs,
        )

    def get_metadef_object(
        self,
        metadef_object: str | _metadef_object.MetadefObject,
        namespace: str | _metadef_namespace.MetadefNamespace,
    ) -> _metadef_object.MetadefObject:
        """Get a single metadef object

        :param metadef_object: The value can be the ID of a metadef_object
            or a
            :class:`~openstack.image.v2.metadef_object.MetadefObject`
            instance.
        :param namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :returns: One :class:`~openstack.image.v2.metadef_object.MetadefObject`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        object_name = resource.Resource._get_id(metadef_object)
        namespace_name = resource.Resource._get_id(namespace)
        return self._get(
            _metadef_object.MetadefObject,
            namespace_name=namespace_name,
            name=object_name,
        )

    def metadef_objects(
        self,
        namespace: str | _metadef_namespace.MetadefNamespace,
    ) -> Generator[_metadef_object.MetadefObject, None, None]:
        """Get metadef object list of the namespace

        :param namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.

        :returns: One :class:`~openstack.image.v2.metadef_object.MetadefObject`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        namespace_name = resource.Resource._get_id(namespace)
        return self._list(
            _metadef_object.MetadefObject,
            namespace_name=namespace_name,
        )

    def update_metadef_object(
        self,
        metadef_object: str | _metadef_object.MetadefObject,
        namespace: str | _metadef_namespace.MetadefNamespace,
        **attrs: Any,
    ) -> _metadef_object.MetadefObject:
        """Update a single metadef object

        :param metadef_object: The value can be the ID of a metadef_object or a
            :class:`~openstack.image.v2.metadef_object.MetadefObject` instance.
        :param namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :param attrs: Keyword arguments which will be used to update
            a :class:`~openstack.image.v2.metadef_object.MetadefObject`

        :returns: One :class:`~openstack.image.v2.metadef_object.MetadefObject`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        namespace_name = resource.Resource._get_id(namespace)
        metadef_object = resource.Resource._get_id(metadef_object)
        return self._update(
            _metadef_object.MetadefObject,
            metadef_object,
            namespace_name=namespace_name,
            **attrs,
        )

    # TODO(stephenfin): This method should return None
    def delete_metadef_object(
        self,
        metadef_object: str | _metadef_object.MetadefObject,
        namespace: str | _metadef_namespace.MetadefNamespace,
        **attrs: Any,
    ) -> _metadef_object.MetadefObject | None:
        """Removes a single metadef object

        :param metadef_object: The value can be the ID of a metadef_object or a
            :class:`~openstack.image.v2.metadef_object.MetadefObject` instance.
        :param namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :param attrs: Keyword arguments which will be used to update
            a :class:`~openstack.image.v2.metadef_object.MetadefObject`

        :returns: The deleted metadef object.
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        namespace_name = resource.Resource._get_id(namespace)
        return self._delete(
            _metadef_object.MetadefObject,
            metadef_object,
            namespace_name=namespace_name,
            **attrs,
        )

    # TODO(stephenfin): This method should return None
    def delete_all_metadef_objects(
        self, namespace: str | _metadef_namespace.MetadefNamespace
    ) -> _metadef_namespace.MetadefNamespace:
        """Delete all objects

        :param namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.
        :returns: The namespace whose objects were deleted.
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        namespace = self._get_resource(
            _metadef_namespace.MetadefNamespace, namespace
        )
        return namespace.delete_all_objects(self)

    # ====== METADEF RESOURCE TYPES ======
    def metadef_resource_types(
        self,
        **query: Any,
    ) -> Generator[_metadef_resource_type.MetadefResourceType, None, None]:
        """Return a generator of metadef resource types

        :returns: A generator object of metadef resource types
            :class:`~openstack.image.v2.metadef_resource_type.MetadefResourceType`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._list(_metadef_resource_type.MetadefResourceType, **query)

    # ====== METADEF RESOURCE TYPES ASSOCIATION======
    def create_metadef_resource_type_association(
        self,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        **attrs: Any,
    ) -> _metadef_resource_type.MetadefResourceTypeAssociation:
        """Creates a resource type association between a namespace
            and the resource type specified in the body of the request.

        :param attrs: Keyword arguments which will be used to create a
            :class:`~openstack.image.v2.metadef_resource_type.MetadefResourceTypeAssociation`
            comprised of the properties on the
            MetadefResourceTypeAssociation class.

        :returns: The results of metadef resource type association creation
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        return self._create(
            _metadef_resource_type.MetadefResourceTypeAssociation,
            namespace_name=namespace_name,
            **attrs,
        )

    def delete_metadef_resource_type_association(
        self,
        metadef_resource_type: str
        | _metadef_resource_type.MetadefResourceTypeAssociation,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        ignore_missing: bool = True,
    ) -> None:
        """Removes a resource type association in a namespace.

        :param metadef_resource_type: The value can be either the name of
            a metadef resource type association or an
            :class:`~openstack.image.v2.metadef_resource_type.MetadefResourceTypeAssociation`
            instance.
        :param metadef_namespace: The value can be either the name of metadef
            namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance
        :param ignore_missing: When set to ``False``,
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the metadef resource type association does not exist.
        :returns: ``None``
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        self._delete(
            _metadef_resource_type.MetadefResourceTypeAssociation,
            metadef_resource_type,
            namespace_name=namespace_name,
            ignore_missing=ignore_missing,
        )

    def metadef_resource_type_associations(
        self,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        **query: Any,
    ) -> Generator[
        _metadef_resource_type.MetadefResourceTypeAssociation, None, None
    ]:
        """Return a generator of metadef resource type associations

        :param metadef_namespace: The value can be either the name of metadef
            namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance
        :returns: A generator object of metadef resource type associations
            :class:`~openstack.image.v2.metadef_resource_type.MetadefResourceTypeAssociation`
        :raises: :class:`~openstack.exceptions.NotFoundException`
                when no resource can be found.
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        return self._list(
            _metadef_resource_type.MetadefResourceTypeAssociation,
            namespace_name=namespace_name,
            **query,
        )

    # ====== METADEF PROPERTY ======
    def create_metadef_property(
        self,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        **attrs: Any,
    ) -> _metadef_property.MetadefProperty:
        """Create a metadef property

        :param metadef_namespace: The value can be either the name of metadef
            namespace or an
            :class:`~openstack.image.v2.metadef_property.MetadefNamespace`
            instance
        :param attrs: The attributes to create on the metadef property
            represented by ``metadef_property``.

        :returns: The created metadef property
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        return self._create(
            _metadef_property.MetadefProperty,
            namespace_name=namespace_name,
            **attrs,
        )

    def update_metadef_property(
        self,
        metadef_property: str | _metadef_property.MetadefProperty,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        **attrs: Any,
    ) -> _metadef_property.MetadefProperty:
        """Update a metadef property

        :param metadef_property: The value can be either the name of metadef
            property or an
            :class:`~openstack.image.v2.metadef_property.MetadefProperty`
            instance.
        :param metadef_namespace: The value can be either the name of metadef
            namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance
        :param attrs: The attributes to update on the metadef property
            represented by ``metadef_property``.

        :returns: The updated metadef property
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        metadef_property = resource.Resource._get_id(metadef_property)
        return self._update(
            _metadef_property.MetadefProperty,
            metadef_property,
            namespace_name=namespace_name,
            **attrs,
        )

    def delete_metadef_property(
        self,
        metadef_property: str | _metadef_property.MetadefProperty,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        ignore_missing: bool = True,
    ) -> None:
        """Delete a metadef property

        :param metadef_property: The value can be either the name of metadef
            property or an
            :class:`~openstack.image.v2.metadef_property.MetadefProperty`
            instance
        :param metadef_namespace: The value can be either the name of metadef
            namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance
        :param ignore_missing: When set to
            ``False`` :class:`~openstack.exceptions.NotFoundException` will be
            raised when the instance does not exist. When set to ``True``,
            no exception will be set when attempting to delete a nonexistent
            instance.

        :returns: ``None``
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        metadef_property = resource.Resource._get_id(metadef_property)
        self._delete(
            _metadef_property.MetadefProperty,
            metadef_property,
            namespace_name=namespace_name,
            ignore_missing=ignore_missing,
        )

    def metadef_properties(
        self,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        **query: Any,
    ) -> Generator[_metadef_property.MetadefProperty, None, None]:
        """Return a generator of metadef properties

        :param metadef_namespace: The value can be either the name of metadef
            namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance
        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of property objects
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        return self._list(
            _metadef_property.MetadefProperty,
            requires_id=False,
            namespace_name=namespace_name,
            **query,
        )

    def get_metadef_property(
        self,
        metadef_property: str | _metadef_property.MetadefProperty,
        metadef_namespace: str | _metadef_namespace.MetadefNamespace,
        **query: Any,
    ) -> _metadef_property.MetadefProperty:
        """Get a single metadef property

        :param metadef_property: The value can be either the name of metadef
            property or an
            :class:`~openstack.image.v2.metadef_property.MetadefProperty`
            instance.
        :param metadef_namespace: The value can be either the name of metadef
            namespace or an
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance

        :returns: One
            :class:`~~openstack.image.v2.metadef_property.MetadefProperty`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        namespace_name = resource.Resource._get_id(metadef_namespace)
        return self._get(
            _metadef_property.MetadefProperty,
            metadef_property,
            namespace_name=namespace_name,
            **query,
        )

    def delete_all_metadef_properties(
        self, metadef_namespace: str | _metadef_namespace.MetadefNamespace
    ) -> None:
        """Delete all metadata definitions property inside a namespace.

        :param metadef_namespace: The value can be either the name of a metadef
            namespace or a
            :class:`~openstack.image.v2.metadef_namespace.MetadefNamespace`
            instance.

        :returns: ``None``
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            resource can be found.
        """
        namespace = self._get_resource(
            _metadef_namespace.MetadefNamespace, metadef_namespace
        )
        namespace.delete_all_properties(self)

    # ====== SCHEMAS ======
    def get_images_schema(self) -> _schema.Schema:
        """Get images schema

        :returns: One :class:`~openstack.image.v2.schema.Schema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _schema.Schema,
            requires_id=False,
            base_path='/schemas/images',
        )

    def get_image_schema(self) -> _schema.Schema:
        """Get single image schema

        :returns: One :class:`~openstack.image.v2.schema.Schema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _schema.Schema,
            requires_id=False,
            base_path='/schemas/image',
        )

    def get_members_schema(self) -> _schema.Schema:
        """Get image members schema

        :returns: One :class:`~openstack.image.v2.schema.Schema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _schema.Schema,
            requires_id=False,
            base_path='/schemas/members',
        )

    def get_member_schema(self) -> _schema.Schema:
        """Get image member schema

        :returns: One :class:`~openstack.image.v2.schema.Schema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _schema.Schema,
            requires_id=False,
            base_path='/schemas/member',
        )

    def get_tasks_schema(self) -> _schema.Schema:
        """Get image tasks schema

        :returns: One :class:`~openstack.image.v2.schema.Schema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _schema.Schema,
            requires_id=False,
            base_path='/schemas/tasks',
        )

    def get_task_schema(self) -> _schema.Schema:
        """Get image task schema

        :returns: One :class:`~openstack.image.v2.schema.Schema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _schema.Schema,
            requires_id=False,
            base_path='/schemas/task',
        )

    def get_metadef_namespace_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition namespace schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/namespace',
        )

    def get_metadef_namespaces_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition namespaces schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/namespaces',
        )

    def get_metadef_resource_type_schema(
        self,
    ) -> _metadef_schema.MetadefSchema:
        """Get metadata definition resource type association schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/resource_type',
        )

    def get_metadef_resource_types_schema(
        self,
    ) -> _metadef_schema.MetadefSchema:
        """Get metadata definition resource type associations schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/resource_types',
        )

    def get_metadef_object_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition object schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/object',
        )

    def get_metadef_objects_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition objects schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/objects',
        )

    def get_metadef_property_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition property schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/property',
        )

    def get_metadef_properties_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition properties schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/properties',
        )

    def get_metadef_tag_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition tag schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/tag',
        )

    def get_metadef_tags_schema(self) -> _metadef_schema.MetadefSchema:
        """Get metadata definition tags schema

        :returns: One :class:`~openstack.image.v2.metadef_schema.MetadefSchema`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(
            _metadef_schema.MetadefSchema,
            requires_id=False,
            base_path='/schemas/metadefs/tags',
        )

    # ====== TASKS ======
    def tasks(self, **query: Any) -> Generator[_task.Task, None, None]:
        """Return a generator of tasks

        :param query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of task objects
        """
        return self._list(_task.Task, **query)

    def get_task(self, task: str | _task.Task) -> _task.Task:
        """Get task details

        :param task: The value can be the ID of a task or a
            :class:`~openstack.image.v2.task.Task` instance.

        :returns: One :class:`~openstack.image.v2.task.Task`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_task.Task, task)

    def create_task(self, **attrs: Any) -> _task.Task:
        """Create a new task from attributes

        :param attrs: Keyword arguments which will be used to create
            a :class:`~openstack.image.v2.task.Task`,
            comprised of the properties on the Task class.

        :returns: The results of task creation
        """
        return self._create(_task.Task, **attrs)

    def wait_for_task(
        self,
        task: _task.Task,
        status: str = 'success',
        failures: list[str] | None = None,
        interval: int | float = 2,
        wait: int | None = 120,
    ) -> _task.Task:
        """Wait for a task to be in a particular status.

        :param task: The resource to wait on to reach the specified status.
            The resource must have a ``status`` attribute.
        :type resource: A :class:`~openstack.resource.Resource` object.
        :param status: Desired status.
        :param failures: Statuses that would be interpreted as failures.
        :type failures: :py:class:`list`
        :param interval: Number of seconds to wait before to consecutive
            checks. Default to 2.
        :param wait: Maximum number of seconds to wait before the change.
            Default to 120.
        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to the desired status failed to occur in specified seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            has transited to one of the failure statuses.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute.
        """
        if failures is None:
            failures = ['failure']
        else:
            failures = [f.lower() for f in failures]

        if task.status.lower() == status.lower():
            return task

        name = f"{task.__class__.__name__}:{task.id}"
        msg = f"Timeout waiting for {name} to transition to {status}"

        for count in utils.iterate_timeout(
            timeout=wait, message=msg, wait=interval
        ):
            task = task.fetch(self)

            if not task:
                raise exceptions.ResourceFailure(
                    f"{name} went away while waiting for {status}"
                )

            new_status = task.status
            normalized_status = new_status.lower()
            if normalized_status == status.lower():
                return task
            elif normalized_status in failures:
                if task.message == _IMAGE_ERROR_396:
                    task_args = {'input': task.input, 'type': task.type}
                    task = self.create_task(**task_args)
                    self.log.debug(f'Got error 396. Recreating task {task}')
                else:
                    raise exceptions.ResourceFailure(
                        f"{name} transitioned to failure state {new_status}"
                    )

            self.log.debug(
                'Still waiting for resource %s to reach state %s, '
                'current state is %s',
                name,
                status,
                new_status,
            )

        return task

    # ====== STORES ======
    def stores(
        self,
        details: bool = False,
        **query: Any,
    ) -> Generator[_si.Store, None, None]:
        """Return a generator of supported image stores

        :returns: A generator of store objects
        """
        if details:
            query['base_path'] = utils.urljoin(_si.Store.base_path, 'detail')
        return self._list(_si.Store, **query)

    # ====== IMPORTS ======
    def get_import_info(self) -> _si.Import:
        """Get a info about image constraints

        :returns: One :class:`~openstack.image.v2.service_info.Import`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_si.Import, requires_id=False)

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :returns: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int | float | None = 2,
        wait: int | None = 120,
        callback: Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)

    def _get_cleanup_dependencies(
        self,
    ) -> dict[str, CleanupDependency] | None:
        return {'image': {'before': ['identity']}}  # type: ignore[typeddict-item]

    def _service_cleanup(
        self,
        dry_run: bool = True,
        client_status_queue: queue.Queue[resource.Resource] | None = None,
        identified_resources: dict[str, resource.Resource] | None = None,
        filters: dict[str, Any] | None = None,
        resource_evaluation_fn: Callable[
            [
                resource.Resource,
                dict[str, Any] | None,
                dict[str, resource.Resource] | None,
            ],
            bool,
        ]
        | None = None,
        skip_resources: Sequence[str] | None = None,
    ) -> None:
        if self.should_skip_resource_cleanup("image", skip_resources):
            return

        project_id = self.get_project_id()

        # Note that images cannot be deleted when they are still being used
        for obj in self.images(owner=project_id):
            self._service_cleanup_del_res(
                self.delete_image,
                obj,
                dry_run=dry_run,
                client_status_queue=client_status_queue,
                identified_resources=identified_resources,
                filters=filters,
                resource_evaluation_fn=resource_evaluation_fn,
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ShareMapping(resource.Resource):
    resource_key = 'share'
    resources_key = 'shares'
    base_path = '/servers/%(server_id)s/shares'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = False
    allow_delete = True
    allow_list = True

    _max_microversion = '2.97'

    #: The location of the share.
    export_location = resource.Body('export_location', default='')
    #: The ID for the server.
    server_id = resource.URI('server_id')
    #: The ID of the share.
    share_id = resource.Body('share_id', alternate_id=True)
    #: The protocol of the share (e.g. NFS, CEPHFS).
    share_proto = resource.Body('share_proto')
    #: The status of the share attachment.
    status = resource.Body('status')
    #: A tag to identify the share attachment, used as mount point
    #: within the instance. If not provided, the share_id is used.
    tag = resource.Body('tag')
    #: The ID of the share mapping.
    uuid = resource.Body('uuid')



# Copyright 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack import exceptions
from openstack.tests.unit import base
from openstack.tests.unit.cloud import fakes


class TestImageSnapshot(base.TestCase):
    def setUp(self):
        super().setUp()
        self.server_id = str(uuid.uuid4())
        self.image_id = str(uuid.uuid4())
        self.server_name = self.getUniqueString('name')
        self.fake_server = fakes.make_fake_server(
            self.server_id, self.server_name
        )

    def test_create_image_snapshot_wait_until_active_never_active(self):
        snapshot_name = 'test-snapshot'
        pending_image = fakes.make_fake_image(self.image_id, status='pending')
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/servers/{self.server_id}/action',
                    headers=dict(
                        Location='{endpoint}/images/{image_id}'.format(
                            endpoint='https://images.example.com',
                            image_id=self.image_id,
                        )
                    ),
                    validate=dict(
                        json={
                            "createImage": {
                                "name": snapshot_name,
                                "metadata": {},
                            }
                        }
                    ),
                ),
                self.get_glance_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=pending_image,
                ),
            ]
        )

        self.assertRaises(
            exceptions.ResourceTimeout,
            self.cloud.create_image_snapshot,
            snapshot_name,
            dict(id=self.server_id),
            wait=True,
            timeout=0.01,
        )

        # After the fifth call, we just keep polling get images for status.
        # Due to mocking sleep, we have no clue how many times we'll call it.
        self.assert_calls(stop_after=5, do_count=False)

    def test_create_image_snapshot_wait_active(self):
        snapshot_name = 'test-snapshot'
        pending_image = fakes.make_fake_image(self.image_id, status='pending')
        fake_image = fakes.make_fake_image(self.image_id)
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/servers/{self.server_id}/action',
                    headers=dict(
                        Location='{endpoint}/images/{image_id}'.format(
                            endpoint='https://images.example.com',
                            image_id=self.image_id,
                        )
                    ),
                    validate=dict(
                        json={
                            "createImage": {
                                "name": snapshot_name,
                                "metadata": {},
                            }
                        }
                    ),
                ),
                self.get_glance_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=pending_image,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=fake_image,
                ),
            ]
        )
        image = self.cloud.create_image_snapshot(
            'test-snapshot', dict(id=self.server_id), wait=True, timeout=2
        )
        self.assertEqual(image['id'], self.image_id)

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.shared_file_system.v2 import service
from openstack.tests.unit import base


EXAMPLE = {
    "status": "enabled",
    "binary": "manila-share",
    "zone": "nova",
    "host": "manila2@generic1",
    "updated_at": "2021-06-07T13:03:57.000000",
    "state": "up",
    "id": 1,
}


class TestServices(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = {'service': {}}
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.status_code = 200
        self.resp.headers = {}
        self.sess = mock.Mock()
        self.sess.put = mock.Mock(return_value=self.resp)

    def test_basic(self):
        services_resource = service.Service()
        self.assertEqual('services', services_resource.resources_key)
        self.assertEqual('/services', services_resource.base_path)
        self.assertTrue(services_resource.allow_list)
        self.assertFalse(services_resource.allow_create)
        self.assertFalse(services_resource.allow_fetch)
        self.assertTrue(services_resource.allow_commit)
        self.assertFalse(services_resource.allow_delete)
        self.assertFalse(services_resource.allow_head)

        self.assertDictEqual(
            {
                'binary': 'binary',
                'host': 'host',
                'limit': 'limit',
                'state': 'state',
                'status': 'status',
                'marker': 'marker',
                'name': 'binary',
                'availability_zone': 'zone',
            },
            services_resource._query_mapping._mapping,
        )

    def test_make_services(self):
        services_resource = service.Service(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], services_resource.id)
        self.assertEqual(EXAMPLE['status'], services_resource.status)
        self.assertEqual(EXAMPLE['binary'], services_resource.binary)
        self.assertEqual(EXAMPLE['zone'], services_resource.availability_zone)
        self.assertEqual(EXAMPLE['host'], services_resource.host)
        self.assertEqual(EXAMPLE['updated_at'], services_resource.updated_at)
        self.assertEqual(EXAMPLE['state'], services_resource.state)

    def test_enable(self):
        sot = service.Service(**EXAMPLE)

        res = sot.enable(self.sess)
        self.assertIsNotNone(res)

        url = 'services/enable'
        body = {
            'binary': EXAMPLE['binary'],
            'host': EXAMPLE['host'],
        }
        self.sess.put.assert_called_with(
            url, json=body, microversion=self.sess.default_microversion
        )

    def test_disable(self):
        sot = service.Service(**EXAMPLE)

        res = sot.disable(self.sess)
        self.assertIsNotNone(res)

        url = 'services/disable'
        body = {
            'binary': EXAMPLE['binary'],
            'host': EXAMPLE['host'],
        }
        self.sess.put.assert_called_with(
            url, json=body, microversion=self.sess.default_microversion
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack.network.v2 import _base
from openstack import resource


class SecurityGroupsDefaultStatefulness(_base.NetworkResource):
    resource_key = 'security_group_default_statefulness'
    resources_key = 'security_groups_default_statefulness'
    base_path = '/security-groups-default-statefulness'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'id',
        'project_id',
        'stateful',
        'sort_dir',
        'sort_key',
    )

    # Properties
    #: The project to which this setting applies.
    #: ``None`` means system-wide.
    project_id = resource.Body('project_id')
    #: The default value for the ``stateful`` attribute when creating
    #: new security groups.
    #: *Type: bool*
    stateful = resource.Body('stateful', type=bool)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional.identity.v3 import base


class TestService(base.BaseIdentityTest):
    def setUp(self):
        super().setUp()
        self.svc_name = self.getUniqueString("crud-service")
        self.svc_type = self.getUniqueString("crud-type")

    def test_service(self):
        # create a new service
        service = self.admin_identity_client.create_service(
            name=self.svc_name,
            type=self.svc_type,
            description="CRUD workflow test",
            enabled=True,
        )
        svc_id = service.id
        self.addCleanup(
            self.admin_identity_client.delete_service,
            svc_id,
            ignore_missing=True,
        )

        self.assertEqual(self.svc_name, service.name)
        self.assertEqual(self.svc_type, service.type)
        self.assertTrue(service.is_enabled)

        # retrieve details of the created service
        got = self.admin_identity_client.get_service(svc_id)
        self.assertEqual(svc_id, got.id)

        # verify service appears in list
        ids = {s.id for s in self.admin_identity_client.services()}
        self.assertIn(svc_id, ids)

        # find service by name
        found = self.admin_identity_client.find_service(
            self.svc_name, ignore_missing=False
        )
        self.assertIsNotNone(found)
        self.assertEqual(svc_id, found.id)

        # update service attributes
        new_desc = "Updated CRUD test"
        updated = self.admin_identity_client.update_service(
            svc_id, description=new_desc, enabled=False
        )
        self.assertEqual(new_desc, updated.description)
        self.assertFalse(updated.is_enabled)

        # delete the service
        self.admin_identity_client.delete_service(svc_id)
        self.assertIsNone(
            self.admin_identity_client.find_service(
                svc_id, ignore_missing=True
            )
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.compute import compute_service
from openstack.compute.v2 import server
from openstack import format as _format
from openstack import resource
from openstack.test import fakes
from openstack.tests.unit import base


class TestGetFake(base.TestCase):
    def test_generate_fake_resource_one(self):
        res = fakes.generate_fake_resource(resource.Resource)
        self.assertIsInstance(res, resource.Resource)

    def test_generate_fake_resource_list(self):
        res = list(fakes.generate_fake_resources(resource.Resource, 2))
        self.assertEqual(2, len(res))
        self.assertIsInstance(res[0], resource.Resource)

    def test_generate_fake_resource_types(self):
        class Foo(resource.Resource):
            a = resource.Body("a", type=str)
            b = resource.Body("b", type=int)
            c = resource.Body("c", type=bool)
            d = resource.Body("d", type=_format.BoolStr)
            e = resource.Body("e", type=dict)
            f = resource.URI("path")

        class Bar(resource.Resource):
            a = resource.Body("a", type=list, list_type=str)
            b = resource.Body("b", type=list, list_type=dict)
            c = resource.Body("c", type=list, list_type=Foo)

        foo = fakes.generate_fake_resource(Foo)
        self.assertIsInstance(foo.a, str)
        self.assertIsInstance(foo.b, int)
        self.assertIsInstance(foo.c, bool)
        self.assertIsInstance(foo.d, bool)
        self.assertIsInstance(foo.e, dict)
        self.assertIsInstance(foo.f, str)

        bar = fakes.generate_fake_resource(Bar)
        self.assertIsInstance(bar.a, list)
        self.assertEqual(1, len(bar.a))
        self.assertIsInstance(bar.a[0], str)
        self.assertIsInstance(bar.b, list)
        self.assertEqual(1, len(bar.b))
        self.assertIsInstance(bar.b[0], dict)
        self.assertIsInstance(bar.c, list)
        self.assertEqual(1, len(bar.c))
        self.assertIsInstance(bar.c[0], Foo)
        self.assertIsInstance(bar.c[0].a, str)
        self.assertIsInstance(bar.c[0].b, int)
        self.assertIsInstance(bar.c[0].c, bool)
        self.assertIsInstance(bar.c[0].d, bool)
        self.assertIsInstance(bar.c[0].e, dict)
        self.assertIsInstance(bar.c[0].f, str)

    def test_generate_fake_resource_attrs(self):
        class Fake(resource.Resource):
            a = resource.Body("a", type=str)
            b = resource.Body("b", type=str)

        res = fakes.generate_fake_resource(Fake, b="bar")
        self.assertIsInstance(res.a, str)
        self.assertIsInstance(res.b, str)
        self.assertEqual("bar", res.b)

    def test_generate_fake_resource_types_inherit(self):
        class Fake(resource.Resource):
            a = resource.Body("a", type=str)

        class FakeInherit(resource.Resource):
            a = resource.Body("a", type=Fake)

        res = fakes.generate_fake_resource(FakeInherit)
        self.assertIsInstance(res.a, Fake)
        self.assertIsInstance(res.a.a, str)

    def test_unknown_attrs_as_props(self):
        class Fake(resource.Resource):
            properties = resource.Body("properties")
            _store_unknown_attrs_as_properties = True

        res = fakes.generate_fake_resource(Fake)
        self.assertIsInstance(res.properties, dict)


class TestGenerateFakeProxy(base.TestCase):
    def test_invalid_service_raises(self):
        from openstack import proxy as _proxy

        self.assertRaises(ValueError, fakes.generate_fake_proxy, _proxy.Proxy)

    def test_get_returns_resource(self):
        fake_proxy = fakes.generate_fake_proxy(compute_service.ComputeService)
        result = fake_proxy.get_server('some-uuid')
        self.assertIsInstance(result, server.Server)

    def test_list_returns_generator_of_resources(self):
        fake_proxy = fakes.generate_fake_proxy(compute_service.ComputeService)
        results = list(fake_proxy.servers())
        self.assertEqual(1, len(results))
        self.assertIsInstance(results[0], server.Server)

    def test_find_returns_resource(self):
        fake_proxy = fakes.generate_fake_proxy(compute_service.ComputeService)
        result = fake_proxy.find_server('some-name')
        self.assertIsInstance(result, server.Server)

    def test_typo_raises_attribute_error(self):
        fake_proxy = fakes.generate_fake_proxy(compute_service.ComputeService)
        self.assertRaises(AttributeError, getattr, fake_proxy, 'serverssss')

    def test_side_effect_can_be_overridden(self):
        fake_proxy = fakes.generate_fake_proxy(compute_service.ComputeService)
        custom = server.Server(id='fixed-id')
        fake_proxy.get_server.side_effect = lambda *a, **kw: custom
        result = fake_proxy.get_server('some-uuid')
        self.assertEqual('fixed-id', result.id)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# alias because we already had an option named argparse
import argparse as argparse_mod
import collections
import copy
import errno
import json
import os
import re
import sys
from typing import Any, TYPE_CHECKING
import warnings

from keystoneauth1 import adapter
from keystoneauth1 import loading
from keystoneauth1 import session
import platformdirs
import yaml

from openstack import _log
from openstack.config import _util
from openstack.config import cloud_region
from openstack.config import defaults
from openstack.config import vendors
from openstack import exceptions
from openstack import warnings as os_warnings

if TYPE_CHECKING:
    from keystoneauth1.loading._plugins.identity import v3 as v3_loaders
    from keystoneauth1.loading import opts

PLATFORMDIRS = platformdirs.PlatformDirs(
    'openstack', 'OpenStack', multipath=True
)
CONFIG_HOME = PLATFORMDIRS.user_config_dir
CACHE_PATH = PLATFORMDIRS.user_cache_dir

# snaps do set $HOME to something like
# /home/$USER/snap/openstackclients/$SNAP_VERSION
# the real home (usually /home/$USERNAME) is stored in $SNAP_REAL_HOME
# see https://snapcraft.io/docs/environment-variables
SNAP_REAL_HOME = os.getenv('SNAP_REAL_HOME')
if SNAP_REAL_HOME:
    UNIX_CONFIG_HOME = os.path.join(
        os.path.join(SNAP_REAL_HOME, '.config'), 'openstack'
    )
else:
    UNIX_CONFIG_HOME = os.path.join(
        os.path.expanduser(os.path.join('~', '.config')), 'openstack'
    )
UNIX_SITE_CONFIG_HOME = '/etc/openstack'

SITE_CONFIG_HOME = PLATFORMDIRS.site_config_dir

CONFIG_SEARCH_PATH = [
    os.getcwd(),
    CONFIG_HOME,
    UNIX_CONFIG_HOME,
    SITE_CONFIG_HOME,
    UNIX_SITE_CONFIG_HOME,
]
YAML_SUFFIXES = ('.yaml', '.yml')
JSON_SUFFIXES = ('.json',)
CONFIG_FILES = [
    os.path.join(d, 'clouds' + s)
    for d in CONFIG_SEARCH_PATH
    for s in YAML_SUFFIXES + JSON_SUFFIXES
]
SECURE_FILES = [
    os.path.join(d, 'secure' + s)
    for d in CONFIG_SEARCH_PATH
    for s in YAML_SUFFIXES + JSON_SUFFIXES
]
VENDOR_FILES = [
    os.path.join(d, 'clouds-public' + s)
    for d in CONFIG_SEARCH_PATH
    for s in YAML_SUFFIXES + JSON_SUFFIXES
]

BOOL_KEYS = ('insecure', 'cache')
CSV_KEYS = ('auth_methods',)

FORMAT_EXCLUSIONS = frozenset(['password'])


def get_boolean(value: Any) -> bool:
    if value is None:
        return False
    if type(value) is bool:
        return value
    if value.lower() == 'true':
        return True
    return False


def _auth_update(
    old_dict: dict[str, Any], new_dict_source: dict[str, Any]
) -> dict[str, Any]:
    """Like dict.update, except handling the nested dict called auth."""
    new_dict = copy.deepcopy(new_dict_source)
    for k, v in new_dict.items():
        if k == 'auth':
            if k in old_dict:
                old_dict[k].update(v)
            else:
                old_dict[k] = v.copy()
        else:
            old_dict[k] = v
    return old_dict


def _fix_argv(argv: list[str]) -> None:
    # Transform any _ characters in arg names to - so that we don't
    # have to throw billions of compat argparse arguments around all
    # over the place.
    processed = collections.defaultdict(set)
    for index in range(0, len(argv)):
        # If the value starts with '--' and has '-' or '_' in it, then
        # it's worth looking at it
        if re.match('^--.*(_|-)+.*', argv[index]):
            split_args = argv[index].split('=')
            orig = split_args[0]
            new = orig.replace('_', '-')
            if orig != new:
                split_args[0] = new
                argv[index] = "=".join(split_args)
            # Save both for later so we can throw an error about dupes
            processed[new].add(orig)
    overlap: list[str] = []
    for new, old in processed.items():
        if len(old) > 1:
            overlap.extend(old)
    if overlap:
        raise exceptions.ConfigException(
            "The following options were given: '{options}' which contain "
            "duplicates except that one has _ and one has -. There is "
            "no sane way for us to know what you're doing. Remove the "
            "duplicate option and try again".format(options=','.join(overlap))
        )


class OpenStackConfig:
    # These two attribute are to allow os-client-config to plumb in its
    # local versions for backwards compat.
    # They should not be used by anyone else.
    _cloud_region_class = cloud_region.CloudRegion
    _defaults_module = defaults

    #: config_filename is the filename that configuration was loaded from, if
    #: any.
    config_filename: str | None
    #: secure_config_filename is the filename that secure configuration was
    #: loaded from, if any.
    secure_config_filename: str | None
    #: cloud_config contains the combined loaded configuration.
    cloud_config: dict[str, Any]

    def __init__(
        self,
        config_files: list[str] | None = None,
        vendor_files: list[str] | None = None,
        override_defaults: dict[str, Any] | None = None,
        force_ipv4: bool | None = None,
        envvar_prefix: str | None = None,
        secure_files: list[str] | None = None,
        pw_func: cloud_region._PasswordCallback | None = None,
        session_constructor: type[session.Session] | None = None,
        app_name: str | None = None,
        app_version: str | None = None,
        load_yaml_config: bool = True,
        load_envvars: bool = True,
        statsd_host: str | None = None,
        statsd_port: str | None = None,
        statsd_prefix: str | None = None,
        influxdb_config: dict[str, Any] | None = None,
    ):
        self.log = _log.setup_logging('openstack.config')
        self._session_constructor = session_constructor
        self._app_name = app_name
        self._app_version = app_version
        self._load_envvars = load_envvars

        if load_yaml_config:
            # "if config_files" is not sufficient to process empty list
            # Use list() to copy rather than reference the module-level
            # constants, preventing mutation of those shared lists when
            # OS_CLIENT_CONFIG_FILE / OS_CLIENT_SECURE_FILE are prepended.
            if config_files is not None:
                self._config_files = list(config_files)
            else:
                self._config_files = list(CONFIG_FILES)
            if secure_files is not None:
                self._secure_files = list(secure_files)
            else:
                self._secure_files = list(SECURE_FILES)
            if vendor_files is not None:
                self._vendor_files = list(vendor_files)
            else:
                self._vendor_files = list(VENDOR_FILES)
        else:
            self._config_files = []
            self._secure_files = []
            self._vendor_files = []

        config_file_override = self._get_envvar('OS_CLIENT_CONFIG_FILE')
        if config_file_override:
            self._config_files.insert(0, config_file_override)

        secure_file_override = self._get_envvar('OS_CLIENT_SECURE_FILE')
        if secure_file_override:
            self._secure_files.insert(0, secure_file_override)

        self.defaults = self._defaults_module.get_defaults()
        if override_defaults:
            self.defaults.update(override_defaults)

        # First, use a config file if it exists where expected
        config_filename, cloud_config = self._load_config_file()
        if config_filename and cloud_config:
            self._validate_config_file(config_filename, cloud_config)

        secure_config_filename, secure_config = self._load_secure_file()
        if secure_config_filename and secure_config:
            self._validate_config_file(secure_config_filename, secure_config)
            cloud_config = _util.merge_clouds(
                cloud_config or {}, secure_config
            )

        self.config_filename = config_filename
        self.secure_config_filename = secure_config_filename
        if not cloud_config:
            self.cloud_config = {'clouds': {}}
        else:
            self.cloud_config = cloud_config
            if 'clouds' not in self.cloud_config:
                self.cloud_config['clouds'] = {}

        # Save the other config
        self.extra_config = copy.deepcopy(self.cloud_config)
        self.extra_config.pop('clouds', None)

        # Grab ipv6 preference settings from env
        client_config = self.cloud_config.get('client', {})

        if force_ipv4 is not None:
            # If it's passed in to the constructor, honor it.
            self.force_ipv4 = force_ipv4
        else:
            # Get the backwards compat value
            prefer_ipv6 = get_boolean(
                self._get_envvar(
                    'OS_PREFER_IPV6',
                    client_config.get(
                        'prefer_ipv6', client_config.get('prefer-ipv6', True)
                    ),
                )
            )
            force_ipv4 = get_boolean(
                self._get_envvar(
                    'OS_FORCE_IPV4',
                    client_config.get(
                        'force_ipv4', client_config.get('broken-ipv6', False)
                    ),
                )
            )

            self.force_ipv4 = force_ipv4
            if not prefer_ipv6:
                # this will only be false if someone set it explicitly
                # honor their wishes
                self.force_ipv4 = True

        # Next, process environment variables and add them to the mix
        self.envvar_key = self._get_envvar('OS_CLOUD_NAME', 'envvars')
        if self.envvar_key in self.cloud_config['clouds']:
            raise exceptions.ConfigException(
                f'{self.config_filename!r} defines a cloud named '
                f'{self.envvar_key!r}, but OS_CLOUD_NAME is also set to '
                f'{self.envvar_key!r}. '
                f'Please rename either your environment-based cloud, '
                f'or one of your file-based clouds.'
            )

        self.default_cloud = self._get_envvar('OS_CLOUD')

        if load_envvars:
            envvars = self._get_os_environ(envvar_prefix=envvar_prefix)
            if envvars:
                self.cloud_config['clouds'][self.envvar_key] = envvars
                if not self.default_cloud:
                    self.default_cloud = self.envvar_key

        if not self.default_cloud and self.cloud_config['clouds']:
            if len(self.cloud_config['clouds'].keys()) == 1:
                # If there is only one cloud just use it. This matches envvars
                # behavior and allows for much less typing.
                # TODO(mordred) allow someone to mark a cloud as "default" in
                # clouds.yaml.
                # The next/iter thing is for python3 compat where dict.keys
                # returns an iterator but in python2 it's a list.
                self.default_cloud = next(
                    iter(self.cloud_config['clouds'].keys())
                )

        # Finally, fall through and make a cloud that starts with defaults
        # because we need somewhere to put arguments, and there are neither
        # config files or env vars
        if not self.cloud_config['clouds']:
            self.cloud_config = dict(clouds=dict(defaults=dict(self.defaults)))
            self.default_cloud = 'defaults'

        self._cache_auth = False
        self._cache_expiration_time = 0
        self._cache_path = CACHE_PATH
        self._cache_class = 'dogpile.cache.null'
        self._cache_arguments: dict[str, Any] = {}
        self._cache_expirations: dict[str, int] = {}
        self._influxdb_config = {}
        if 'cache' in self.cloud_config:
            cache_settings = _util.normalize_keys(self.cloud_config['cache'])

            self._cache_auth = get_boolean(
                cache_settings.get('auth', self._cache_auth)
            )

            # expiration_time used to be 'max_age' but the dogpile setting
            # is expiration_time. Support max_age for backwards compat.
            self._cache_expiration_time = cache_settings.get(
                'expiration_time',
                cache_settings.get('max_age', self._cache_expiration_time),
            )

            # If cache class is given, use that. If not, but if cache time
            # is given, default to memory. Otherwise, default to nothing.
            if self._cache_expiration_time:
                self._cache_class = 'dogpile.cache.memory'
            self._cache_class = self.cloud_config['cache'].get(
                'class', self._cache_class
            )

            self._cache_path = os.path.expanduser(
                cache_settings.get('path', self._cache_path)
            )
            self._cache_arguments = cache_settings.get(
                'arguments', self._cache_arguments
            )
            self._cache_expirations = cache_settings.get(
                'expiration', self._cache_expirations
            )

        if load_yaml_config:
            metrics_config = self.cloud_config.get('metrics', {})
            statsd_config = metrics_config.get('statsd', {})
            statsd_host = statsd_host or statsd_config.get('host')
            statsd_port = statsd_port or statsd_config.get('port')
            statsd_prefix = statsd_prefix or statsd_config.get('prefix')

            influxdb_cfg = metrics_config.get('influxdb', {})
            # Parse InfluxDB configuration
            if not influxdb_config:
                influxdb_config = influxdb_cfg
            else:
                influxdb_config.update(influxdb_cfg)

        if influxdb_config:
            # NOTE(stephenfin): defer the warning to here so we catch config in
            # both clouds.yaml and directly passed in
            warnings.warn(
                'Support for InfluxDB requires the influxdb library which '
                'only supports InfluxDB 1.x and is deprecated. As a result, '
                'influxdb is also deprecated and will be removed in a future '
                'release.',
                os_warnings.RemovedInSDK60Warning,
            )

            config = {}
            if 'use_udp' in influxdb_config:
                use_udp = influxdb_config['use_udp']
                if isinstance(use_udp, str):
                    use_udp = use_udp.lower() in ('true', 'yes', '1')
                elif not isinstance(use_udp, bool):
                    use_udp = False
                    self.log.warning(
                        'InfluxDB.use_udp value type is not '
                        'supported. Use one of '
                        '[true|false|yes|no|1|0]'
                    )
                config['use_udp'] = use_udp
            for key in [
                'host',
                'port',
                'username',
                'password',
                'database',
                'measurement',
                'timeout',
            ]:
                if key in influxdb_config:
                    config[key] = influxdb_config[key]
            self._influxdb_config = config

        if load_envvars:
            statsd_host = statsd_host or os.environ.get('STATSD_HOST')
            statsd_port = statsd_port or os.environ.get('STATSD_PORT')
            statsd_prefix = statsd_prefix or os.environ.get('STATSD_PREFIX')

        self._statsd_host = statsd_host
        self._statsd_port = statsd_port
        self._statsd_prefix = statsd_prefix

        # Flag location to hold the peeked value of an argparse timeout value
        self._argv_timeout = False

        # Save the password callback
        # password = self._pw_callback(prompt="Password: ")
        self._pw_callback = pw_func

    def _get_os_environ(
        self, envvar_prefix: str | None = None
    ) -> dict[str, Any] | None:
        ret = self._defaults_module.get_defaults()
        if not envvar_prefix:
            # This makes the or below be OS_ or OS_ which is a no-op
            envvar_prefix = 'OS_'
        environkeys = [
            k
            for k in os.environ.keys()
            if (k.startswith('OS_') or k.startswith(envvar_prefix))
            and not k.startswith('OS_TEST')  # infra CI var
            and not k.startswith('OS_STD')  # oslotest var
            and not k.startswith('OS_LOG')  # oslotest var
        ]
        for k in environkeys:
            newkey = k.split('_', 1)[-1].lower()
            ret[newkey] = os.environ[k]
        # If the only environ keys are selectors or behavior modification,
        # don't return anything
        selectors = {
            'OS_CLOUD',
            'OS_REGION_NAME',
            'OS_CLIENT_CONFIG_FILE',
            'OS_CLIENT_SECURE_FILE',
            'OS_CLOUD_NAME',
        }
        if set(environkeys) - selectors:
            return ret
        return None

    def _get_envvar(self, key: str, default: str | None = None) -> str | None:
        if not self._load_envvars:
            return default
        return os.environ.get(key, default)

    def get_extra_config(
        self, key: str, defaults: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """Fetch an arbitrary extra chunk of config, laying in defaults.

        :param string key: name of the config section to fetch
        :param dict defaults: (optional) default values to merge under the
                                         found config
        """
        defaults = _util.normalize_keys(defaults or {})
        assert defaults is not None  # narrow type
        if not key:
            return defaults
        return _util.merge_clouds(
            defaults, _util.normalize_keys(self.cloud_config.get(key, {}))
        )

    def _load_config_file(
        self,
    ) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        return self._load_yaml_json_file(self._config_files)

    def _load_secure_file(
        self,
    ) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        return self._load_yaml_json_file(self._secure_files)

    def _load_vendor_file(
        self,
    ) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        return self._load_yaml_json_file(self._vendor_files)

    def _load_yaml_json_file(
        self, filelist: list[str]
    ) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        for path in filelist:
            if os.path.exists(path):
                try:
                    with open(path) as f:
                        if path.endswith('json'):
                            return path, json.load(f)
                        else:
                            return path, yaml.safe_load(f)
                except OSError as e:
                    if e.errno == errno.EACCES:
                        # Can't access file so let's continue to the next
                        # file
                        continue
                except (ValueError, yaml.YAMLError) as e:
                    self.log.warning('Skipping config file %s: %s', path, e)
                    continue
        return (None, None)

    def _validate_config_file(self, path: str, data: Any) -> bool:
        """Validate config file contains a clouds entry.

        All config files should have a 'clouds' key at a minimum.
        """
        if not isinstance(data, dict):
            raise exceptions.ConfigException(
                f'Configuration file {path} is empty or not a valid mapping'
            )

        if 'clouds' not in data:
            # TODO(stephenfin): This should probably be an error at some point
            self.log.warning(
                "Configuration file %s does not contain a 'clouds' key", path
            )
            return False

        return True

    def _expand_region_name(self, region_name: str) -> dict[str, Any]:
        return {'name': region_name, 'values': {}}

    def _expand_regions(
        self, regions: list[str | dict[str, Any]]
    ) -> list[dict[str, Any]]:
        ret = []
        for region in regions:
            if isinstance(region, dict):
                # i.e. must have name key, and only name,values keys
                if 'name' not in region or not {'name', 'values'} >= set(
                    region
                ):
                    raise exceptions.ConfigException(
                        f'Invalid region entry at: {region}'
                    )
                if 'values' not in region:
                    region['values'] = {}
                ret.append(copy.deepcopy(region))
            else:
                ret.append(self._expand_region_name(region))
        return ret

    def _get_regions(self, cloud: str) -> list[dict[str, Any]]:
        if cloud not in self.cloud_config['clouds']:
            return [self._expand_region_name('')]
        regions = self._get_known_regions(cloud)
        if not regions:
            # We don't know of any regions use a workable default.
            regions = [self._expand_region_name('')]
        return regions

    def _get_known_regions(self, cloud: str) -> list[dict[str, Any]]:
        config = _util.normalize_keys(self.cloud_config['clouds'][cloud])
        if 'regions' in config:
            return self._expand_regions(config['regions'])
        elif 'region_name' in config:
            if isinstance(config['region_name'], list):
                regions = config['region_name']
            else:
                regions = config['region_name'].split(',')
            if len(regions) > 1:
                warnings.warn(
                    f"Comma separated lists in region_name are deprecated. "
                    f"Please use a yaml list in the regions "
                    f"parameter in {self.config_filename} instead.",
                    os_warnings.OpenStackDeprecationWarning,
                )
            return self._expand_regions(regions)
        else:
            # crappit. we don't have a region defined.
            new_cloud: dict[str, Any] = {}
            our_cloud = self.cloud_config['clouds'].get(cloud, {})
            self._expand_vendor_profile(cloud, new_cloud, our_cloud)
            if new_cloud.get('regions'):
                return self._expand_regions(new_cloud['regions'])
            elif new_cloud.get('region_name'):
                return [self._expand_region_name(new_cloud['region_name'])]

        return []

    def _get_region(
        self, cloud: str | None = None, region_name: str = ''
    ) -> dict[str, Any]:
        if region_name is None:
            region_name = ''

        if not cloud:
            return self._expand_region_name(region_name)

        regions = self._get_known_regions(cloud)
        if not regions:
            return self._expand_region_name(region_name)

        if not region_name:
            return regions[0]

        for region in regions:
            if region['name'] == region_name:
                return region

        raise exceptions.ConfigException(
            'Region {region_name} is not a valid region name for cloud '
            '{cloud}. Valid choices are {region_list}. Please note that '
            'region names are case sensitive.'.format(
                region_name=region_name,
                region_list=','.join([r['name'] for r in regions]),
                cloud=cloud,
            )
        )

    def get_cloud_names(self) -> list[str]:
        return list(self.cloud_config['clouds'].keys())

    def _get_base_cloud_config(
        self, name: str | None, profile: str | None = None
    ) -> dict[str, Any]:
        cloud = {}

        # Only validate cloud name if one was given
        if name and name not in self.cloud_config['clouds']:
            raise exceptions.ConfigException(f"Cloud {name} was not found.")

        our_cloud = self.cloud_config['clouds'].get(name, dict())
        if profile:
            our_cloud['profile'] = profile

        # Get the defaults
        cloud.update(self.defaults)
        self._expand_vendor_profile(name, cloud, our_cloud)

        if 'auth' not in cloud:
            cloud['auth'] = {}

        _auth_update(cloud, our_cloud)
        cloud.pop('cloud', None)

        return cloud

    def _expand_vendor_profile(
        self,
        name: str | None,
        cloud: dict[str, Any],
        our_cloud: dict[str, Any],
    ) -> None:
        # Expand a profile if it exists. 'cloud' is an old confusing name
        # for this.
        profile_name = our_cloud.get('profile', our_cloud.get('cloud', None))
        if not profile_name or profile_name == self.envvar_key:
            return
        if 'cloud' in our_cloud:
            warnings.warn(
                f"{self.config_filename} uses the keyword 'cloud' to "
                f"reference a known vendor profile. This has been deprecated "
                f"in favor of the 'profile' keyword.",
                os_warnings.OpenStackDeprecationWarning,
            )

        _, vendor_file = self._load_vendor_file()
        if (
            vendor_file
            and 'public-clouds' in vendor_file
            and profile_name in vendor_file['public-clouds']
        ):
            _auth_update(cloud, vendor_file['public-clouds'][profile_name])
        else:
            profile_data = vendors.get_profile(profile_name)
            if profile_data:
                nested_profile = profile_data.pop('profile', None)
                if nested_profile:
                    nested_profile_data = vendors.get_profile(nested_profile)
                    if nested_profile_data:
                        profile_data = nested_profile_data
                status = profile_data.pop('status', 'active')
                message = profile_data.pop('message', '')
                if status == 'deprecated':
                    warnings.warn(
                        f"{profile_name} is deprecated: {message}",
                        os_warnings.OpenStackDeprecationWarning,
                    )
                elif status == 'shutdown':
                    raise exceptions.ConfigException(
                        f"{profile_name} references a cloud that no longer "
                        f"exists: {message}"
                    )
                _auth_update(cloud, profile_data)
            else:
                # Can't find the requested vendor config, go about business
                warnings.warn(
                    f"Couldn't find the vendor profile {profile_name} for "
                    f"the cloud {name}",
                    os_warnings.ConfigurationWarning,
                )

    def _project_scoped(self, cloud: dict[str, Any]) -> bool:
        return (
            'project_id' in cloud
            or 'project_name' in cloud
            or 'project_id' in cloud['auth']
            or 'project_name' in cloud['auth']
        )

    def _validate_networks(
        self, networks: list[dict[str, Any]], key: str
    ) -> None:
        value = None
        for net in networks:
            if value and net[key]:
                raise exceptions.ConfigException(
                    "Duplicate network entries for {key}: {net1} and {net2}. "
                    "Only one network can be flagged with {key}".format(
                        key=key, net1=value['name'], net2=net['name']
                    )
                )
            if not value and net[key]:
                value = net

    def _fix_backwards_networks(self, cloud: dict[str, Any]) -> dict[str, Any]:
        # Leave the external_network and internal_network keys in the
        # dict because consuming code might be expecting them.
        networks = []
        # Normalize existing network entries
        for net in cloud.get('networks', []):
            name = net.get('name')
            if not name:
                raise exceptions.ConfigException(
                    'Entry in network list is missing required field "name".'
                )
            network = dict(
                name=name,
                routes_externally=get_boolean(net.get('routes_externally')),
                nat_source=get_boolean(net.get('nat_source')),
                nat_destination=get_boolean(net.get('nat_destination')),
                default_interface=get_boolean(net.get('default_interface')),
            )
            # routes_ipv4_externally defaults to the value of routes_externally
            network['routes_ipv4_externally'] = get_boolean(
                net.get('routes_ipv4_externally', network['routes_externally'])
            )
            # routes_ipv6_externally defaults to the value of routes_externally
            network['routes_ipv6_externally'] = get_boolean(
                net.get('routes_ipv6_externally', network['routes_externally'])
            )
            networks.append(network)

        for key in ('external_network', 'internal_network'):
            external = key.startswith('external')
            if key in cloud and 'networks' in cloud:
                raise exceptions.ConfigException(
                    f"Both {key} and networks were specified in the config. "
                    f"Please remove {key} from the config and use the network "
                    f"list to configure network behavior."
                )
            if key in cloud:
                warnings.warn(
                    f"{key} is deprecated. Please replace with an entry in "
                    f"a dict inside of the networks list with name: "
                    f"{cloud[key]} and routes_externally: {external}",
                    os_warnings.OpenStackDeprecationWarning,
                )
                networks.append(
                    dict(
                        name=cloud[key],
                        routes_externally=external,
                        nat_destination=not external,
                        default_interface=external,
                    )
                )

        # Validate that we don't have duplicates
        self._validate_networks(networks, 'nat_destination')
        self._validate_networks(networks, 'default_interface')

        cloud['networks'] = networks
        return cloud

    def _handle_domain_id(self, cloud: dict[str, Any]) -> dict[str, Any]:
        # Allow people to just specify domain once if it's the same
        mappings = {
            'domain_id': ('user_domain_id', 'project_domain_id'),
            'domain_name': ('user_domain_name', 'project_domain_name'),
        }
        for target_key, possible_values in mappings.items():
            if not self._project_scoped(cloud):
                if target_key in cloud and target_key not in cloud['auth']:
                    cloud['auth'][target_key] = cloud.pop(target_key)
                continue
            for key in possible_values:
                if target_key in cloud['auth'] and key not in cloud['auth']:
                    cloud['auth'][key] = cloud['auth'][target_key]
            cloud.pop(target_key, None)
            cloud['auth'].pop(target_key, None)
        return cloud

    def _fix_backwards_auth(self, cloud: dict[str, Any]) -> dict[str, Any]:
        mappings = {
            'domain_id': ('domain_id', 'domain-id'),
            'domain_name': ('domain_name', 'domain-name'),
            'user_domain_id': ('user_domain_id', 'user-domain-id'),
            'user_domain_name': ('user_domain_name', 'user-domain-name'),
            'project_domain_id': ('project_domain_id', 'project-domain-id'),
            'project_domain_name': (
                'project_domain_name',
                'project-domain-name',
            ),
            'token': ('auth-token', 'auth_token', 'token'),
            'passcode': ('passcode',),
        }
        if cloud.get('auth_type', None) == 'v2password':
            # If v2password is explcitly requested, this is to deal with old
            # clouds. That's fine - we need to map settings in the opposite
            # direction
            mappings['tenant_id'] = (
                'project_id',
                'project-id',
                'tenant_id',
                'tenant-id',
            )
            mappings['tenant_name'] = (
                'project_name',
                'project-name',
                'tenant_name',
                'tenant-name',
            )
        else:
            mappings['project_id'] = (
                'tenant_id',
                'tenant-id',
                'project_id',
                'project-id',
            )
            mappings['project_name'] = (
                'tenant_name',
                'tenant-name',
                'project_name',
                'project-name',
            )
        for target_key, possible_values in mappings.items():
            target = None
            for key in possible_values:
                if key in cloud['auth']:
                    target = str(cloud['auth'][key])
                    del cloud['auth'][key]
                # Prefer values NOT from the 'auth' section
                # as they may contain cli or environment overrides.
                # See story 2010784 for context.
                if key in cloud:
                    target = str(cloud[key])
                    del cloud[key]
            if target:
                cloud['auth'][target_key] = target
        return cloud

    def _fix_backwards_auth_plugin(
        self, cloud: dict[str, Any]
    ) -> dict[str, Any]:
        # Do the lists backwards so that auth_type is the ultimate winner
        mappings = {
            'auth_type': ('auth_plugin', 'auth_type'),
        }
        for target_key, possible_values in mappings.items():
            target = None
            for key in possible_values:
                if key in cloud:
                    target = cloud[key]
                    del cloud[key]
            cloud[target_key] = target
        # Because we force alignment to v3 nouns, we want to force
        # use of the auth plugin that can do auto-selection and dealing
        # with that based on auth parameters. v2password is basically
        # completely broken
        return cloud

    def register_argparse_arguments(
        self,
        parser: argparse_mod.ArgumentParser,
        argv: list[str],
        service_keys: list[str] | None = None,
    ) -> None:
        """Register all of the common argparse options needed.

        Given an argparse parser, register the keystoneauth Session arguments,
        the keystoneauth Auth Plugin Options and os-cloud. Also, peek in the
        argv to see if all of the auth plugin options should be registered
        or merely the ones already configured.

        :param argparse.ArgumentParser: parser to attach argparse options to
        :param argv: the arguments provided to the application
        :param string service_keys: Service or list of services this argparse
                                    should be specialized for, if known.
                                    The first item in the list will be used
                                    as the default value for service_type
                                    (optional)

        :raises exceptions.ConfigException if an invalid auth-type is requested
        """

        if service_keys is None:
            service_keys = []

        # Fix argv in place - mapping any keys with embedded _ in them to -
        _fix_argv(argv)

        local_parser = argparse_mod.ArgumentParser(add_help=False)

        for p in (parser, local_parser):
            p.add_argument(
                '--os-cloud',
                metavar='<name>',
                default=self._get_envvar('OS_CLOUD', None),
                help='Named cloud to connect to',
            )

        # we need to peek to see if timeout was actually passed, since
        # the keystoneauth declaration of it has a default, which means
        # we have no clue if the value we get is from the ksa default
        # for from the user passing it explicitly. We'll stash it for later
        local_parser.add_argument('--timeout', metavar='<timeout>')

        # We need for get_one to be able to peek at whether a token
        # was passed so that we can swap the default from password to
        # token if it was. And we need to also peek for --os-auth-token
        # for novaclient backwards compat
        local_parser.add_argument('--os-token')
        local_parser.add_argument('--os-auth-token')

        # Peek into the future and see if we have an auth-type set in
        # config AND a cloud set, so that we know which command line
        # arguments to register and show to the user (the user may want
        # to say something like:
        #   openstack --os-cloud=foo --os-oidctoken=bar
        # although I think that user is the cause of my personal pain
        options, _args = local_parser.parse_known_args(argv)
        if options.timeout:
            self._argv_timeout = True

        # validate = False because we're not _actually_ loading here
        # we're only peeking, so it's the wrong time to assert that
        # the rest of the arguments given are invalid for the plugin
        # chosen (for instance, --help may be requested, so that the
        # user can see what options he may want to give
        cloud_region = self.get_one(argparse=options, validate=False)
        default_auth_type = cloud_region.config['auth_type']

        try:
            loading.register_auth_argparse_arguments(
                parser, argv, default=default_auth_type
            )
        except Exception:
            # Hidiing the keystoneauth exception because we're not actually
            # loading the auth plugin at this point, so the error message
            # from it doesn't actually make sense to os-client-config users
            options, _args = parser.parse_known_args(argv)
            plugin_names = loading.get_available_plugin_names()
            raise exceptions.ConfigException(
                "An invalid auth-type was specified: {auth_type}. "
                "Valid choices are: {plugin_names}.".format(
                    auth_type=options.os_auth_type,
                    plugin_names=",".join(plugin_names),
                )
            )

        if service_keys:
            primary_service = service_keys[0]
        else:
            primary_service = None
        loading.register_session_argparse_arguments(parser)
        adapter.register_adapter_argparse_arguments(
            parser, service_type=primary_service
        )
        for service_key in service_keys:
            # legacy clients have un-prefixed api-version options
            parser.add_argument(
                '--{service_key}-api-version'.format(
                    service_key=service_key.replace('_', '-')
                ),
                help=argparse_mod.SUPPRESS,
            )
            adapter.register_service_adapter_argparse_arguments(
                parser, service_type=service_key
            )

        # Backwards compat options for legacy clients
        parser.add_argument('--http-timeout', help=argparse_mod.SUPPRESS)
        parser.add_argument('--os-endpoint-type', help=argparse_mod.SUPPRESS)
        parser.add_argument('--endpoint-type', help=argparse_mod.SUPPRESS)

    def _fix_backwards_interface(
        self, cloud: dict[str, Any]
    ) -> dict[str, Any]:
        new_cloud = {}
        for key in cloud.keys():
            if key.endswith('endpoint_type'):
                target_key = key.replace('endpoint_type', 'interface')
            else:
                target_key = key
            new_cloud[target_key] = cloud[key]
        return new_cloud

    def _fix_backwards_api_timeout(
        self, cloud: dict[str, Any]
    ) -> dict[str, Any]:
        new_cloud = {}
        # requests can only have one timeout, which means that in a single
        # cloud there is no point in different timeout values. However,
        # for some reason many of the legacy clients decided to shove their
        # service name in to the arg name for reasons surpassin sanity. If
        # we find any values that are not api_timeout, overwrite api_timeout
        # with the value
        service_timeout = None
        for key in cloud.keys():
            if key.endswith('timeout') and not (
                key == 'timeout' or key == 'api_timeout'
            ):
                service_timeout = cloud[key]
            else:
                new_cloud[key] = cloud[key]
        if service_timeout is not None:
            new_cloud['api_timeout'] = service_timeout
        # The common argparse arg from keystoneauth is called timeout, but
        # os-client-config expects it to be called api_timeout
        if self._argv_timeout:
            if new_cloud.get('timeout'):
                new_cloud['api_timeout'] = new_cloud.pop('timeout')
        return new_cloud

    def get_all(self) -> list[cloud_region.CloudRegion]:
        clouds = []

        for cloud in self.get_cloud_names():
            for region in self._get_regions(cloud):
                if region:
                    clouds.append(
                        self.get_one(cloud, region_name=region['name'])
                    )
        return clouds

    def get_all_clouds(self) -> list[cloud_region.CloudRegion]:
        warnings.warn(
            "The 'get_all_clouds' method is a deprecated alias for "
            "'get_clouds' and will be removed in a future release.",
            os_warnings.RemovedInSDK60Warning,
        )
        return self.get_all()

    def _fix_args(
        self,
        args: dict[str, Any] | None = None,
        argparse: argparse_mod.Namespace | None = None,
    ) -> dict[str, Any]:
        """Massage the passed-in options

        Replace - with _ and strip os_ prefixes.

        Convert an argparse Namespace object to a dict, removing values
        that are either None or ''.
        """
        if not args:
            args = {}

        if argparse:
            # Convert the passed-in Namespace
            o_dict = vars(argparse)
            parsed_args = dict()
            for k in o_dict:
                if o_dict[k] is not None and o_dict[k] != '':
                    parsed_args[k] = o_dict[k]
            args.update(parsed_args)

        os_args = {}
        new_args = {}
        for key, val in iter(args.items()):
            if isinstance(args[key], dict):
                # dive into the auth dict
                new_args[key] = self._fix_args(args[key])
                continue

            key = key.replace('-', '_')
            if key.startswith('os_'):
                os_args[key[3:]] = val
            else:
                new_args[key] = val
        new_args.update(os_args)
        return new_args

    def _find_winning_auth_value(
        self, opt: 'opts.Opt', config: dict[str, dict[str, Any]]
    ) -> dict[str, Any] | None:
        opt_name = opt.name.replace('-', '_')
        if opt_name in config:
            return config[opt_name]
        else:
            deprecated = getattr(
                opt, 'deprecated', getattr(opt, 'deprecated_opts', [])
            )
            for d_opt in deprecated:
                d_opt_name = d_opt.name.replace('-', '_')
                if d_opt_name in config:
                    return config[d_opt_name]

        return None

    def auth_config_hook(self, config: dict[str, Any]) -> dict[str, Any]:
        """Allow examination of config values before loading auth plugin

        OpenStackClient will override this to perform additional checks
        on auth_type.
        """
        return config

    def _get_auth_loader(
        self, config: dict[str, Any]
    ) -> loading.BaseLoader[Any]:
        # Use the 'none' plugin for variants of None specified,
        # since it does not look up endpoints or tokens but rather
        # does a passthrough. This is useful for things like Ironic
        # that have a keystoneless operational mode, but means we're
        # still dealing with a keystoneauth Session object, so all the
        # _other_ things (SSL arg handling, timeout) all work consistently
        if config['auth_type'] in (None, "None", ''):
            config['auth_type'] = 'none'
        elif config['auth_type'] == 'token_endpoint':
            # Humans have been trained to use a thing called token_endpoint
            # That it does not exist in keystoneauth is irrelvant- it not
            # doing what they want causes them sorrow.
            config['auth_type'] = 'admin_token'

        loader: loading.BaseLoader[Any] = loading.get_plugin_loader(
            config['auth_type']
        )

        # As the name would suggest, v3multifactor uses multiple factors for
        # authentication. As a result, we need to register the configuration
        # options for each required auth method. Normally, this is handled by
        # the 'MultiFactor.load_from_options' method but there doesn't appear
        # to be a way to "register" the auth methods without actually loading
        # the plugin. As a result, if we encounter this auth type then we need
        # to do this registration of extra options manually.
        # FIXME(stephenfin): We need to provide a mechanism to extend the
        # options in keystoneauth1.loading._plugins.identity.v3.MultiAuth
        # without calling 'load_from_options'.
        if config['auth_type'] == 'v3multifactor':
            if TYPE_CHECKING:
                # narrow types
                assert isinstance(loader, v3_loaders.MultiFactor)
            # We use '.get' since we can't be sure this key is set yet -
            # validation happens later, in _validate_auth
            loader._methods = config.get('auth_methods')

        return loader

    def _validate_auth(
        self, config: dict[str, Any], loader: loading.BaseLoader[Any]
    ) -> dict[str, Any]:
        # May throw a keystoneauth1.exceptions.NoMatchingPlugin

        plugin_options = loader.get_options()

        for p_opt in plugin_options:
            # if it's in config.auth, win, kill it from config dict
            # if it's in config and not in config.auth, move it
            # deprecated loses to current
            # provided beats default, deprecated or not
            winning_value = self._find_winning_auth_value(
                p_opt,
                config['auth'],
            )
            if not winning_value:
                winning_value = self._find_winning_auth_value(
                    p_opt,
                    config,
                )

            config = self._clean_up_after_ourselves(
                config,
                p_opt,
                winning_value,
            )

            if winning_value:
                # Prefer the plugin configuration dest value if the value's key
                # is marked as deprecated.
                if p_opt.dest is None:
                    good_name = p_opt.name.replace('-', '_')
                    config['auth'][good_name] = winning_value
                else:
                    config['auth'][p_opt.dest] = winning_value

            # See if this needs a prompting
            config = self.option_prompt(config, p_opt)

        return config

    def _validate_auth_correctly(
        self, config: dict[str, Any], loader: loading.BaseLoader[Any]
    ) -> dict[str, Any]:
        # May throw a keystoneauth1.exceptions.NoMatchingPlugin

        plugin_options = loader.get_options()

        for p_opt in plugin_options:
            # if it's in config, win, move it and kill it from config dict
            # if it's in config.auth but not in config it's good
            # deprecated loses to current
            # provided beats default, deprecated or not
            winning_value = self._find_winning_auth_value(
                p_opt,
                config,
            )
            if not winning_value:
                winning_value = self._find_winning_auth_value(
                    p_opt,
                    config['auth'],
                )

            config = self._clean_up_after_ourselves(
                config,
                p_opt,
                winning_value,
            )

            # See if this needs a prompting
            config = self.option_prompt(config, p_opt)

        return config

    def option_prompt(
        self, config: dict[str, Any], p_opt: 'opts.Opt'
    ) -> dict[str, Any]:
        """Prompt user for option that requires a value"""
        if (
            getattr(p_opt, 'prompt', None) is not None
            and p_opt.dest not in config['auth']
            and self._pw_callback is not None
        ):
            config['auth'][p_opt.dest] = self._pw_callback(p_opt.prompt)
        return config

    def _clean_up_after_ourselves(
        self,
        config: dict[str, Any],
        p_opt: 'opts.Opt',
        winning_value: Any,
    ) -> dict[str, Any]:
        # Clean up after ourselves
        for opt in [p_opt.name] + [o.name for o in p_opt.deprecated]:
            opt = opt.replace('-', '_')
            config.pop(opt, None)
            config['auth'].pop(opt, None)

        if winning_value:
            # Prefer the plugin configuration dest value if the value's key
            # is marked as depreciated.
            if p_opt.dest is None:
                config['auth'][p_opt.name.replace('-', '_')] = winning_value
            else:
                config['auth'][p_opt.dest] = winning_value
        return config

    def _handle_value_types(self, config: dict[str, Any]) -> dict[str, Any]:
        for key in BOOL_KEYS:
            if key in config:
                if not isinstance(config[key], bool):
                    config[key] = get_boolean(config[key])

        for key in CSV_KEYS:
            if key in config:
                if isinstance(config[key], str):
                    config[key] = config[key].split(',')
        return config

    def magic_fixes(self, config: dict[str, Any]) -> dict[str, Any]:
        """Perform the set of magic argument fixups"""

        # These backwards compat values are only set via argparse. If it's
        # there, it's because it was passed in explicitly, and should win
        config = self._fix_backwards_api_timeout(config)
        config = self._fix_backwards_auth_plugin(config)
        config = self._fix_backwards_auth(config)
        config = self._fix_backwards_interface(config)
        config = self._fix_backwards_networks(config)
        config = self._handle_domain_id(config)
        config = self._handle_value_types(config)

        # TODO(mordred): Special casing auth_url here. We should
        #                come back to this betterer later so that it's
        #                more generalized
        if 'auth' in config and 'auth_url' in config['auth']:
            config['auth']['auth_url'] = config['auth']['auth_url'].format(
                **config
            )

        return config

    def get_one(
        self,
        cloud: str | None = None,
        validate: bool = True,
        argparse: argparse_mod.Namespace | None = None,
        **kwargs: Any,
    ) -> cloud_region.CloudRegion:
        """Retrieve a single CloudRegion and merge additional options

        :param string cloud:
            The name of the configuration to load from clouds.yaml
        :param boolean validate:
            Validate the config. Setting this to False causes no auth plugin
            to be created. It's really only useful for testing.
        :param Namespace argparse:
            An argparse Namespace object; allows direct passing in of
            argparse options to be added to the cloud config.  Values
            of None and '' will be removed.
        :param region_name: Name of the region of the cloud.
        :param kwargs: Additional configuration options

        :returns: openstack.config.cloud_region.CloudRegion
        :raises: keystoneauth1.exceptions.MissingRequiredOptions
            on missing required auth parameters
        """

        profile = kwargs.pop('profile', None)
        args = self._fix_args(kwargs, argparse=argparse)

        if cloud is None:
            if 'cloud' in args:
                cloud = args['cloud']
            else:
                cloud = self.default_cloud

        config = self._get_base_cloud_config(cloud, profile)

        # Get region specific settings
        if 'region_name' not in args:
            args['region_name'] = ''
        region = self._get_region(cloud=cloud, region_name=args['region_name'])
        args['region_name'] = region['name']
        region_args = copy.deepcopy(region['values'])

        # Regions is a list that we can use to create a list of cloud/region
        # objects. It does not belong in the single-cloud dict
        config.pop('regions', None)

        # Can't just do update, because None values take over
        for arg_list in region_args, args:
            for key, val in iter(arg_list.items()):
                if val is not None:
                    if key == 'auth' and config[key] is not None:
                        config[key] = _auth_update(config[key], val)
                    else:
                        config[key] = val

        config = self.magic_fixes(config)
        config = _util.normalize_keys(config)

        # NOTE(dtroyer): OSC needs a hook into the auth args before the
        #                plugin is loaded in order to maintain backward-
        #                compatible behaviour
        config = self.auth_config_hook(config)

        if validate:
            loader = self._get_auth_loader(config)
            config = self._validate_auth(config, loader)
            auth_plugin = loader.load_from_options(**config['auth'])
        else:
            auth_plugin = None

        # If any of the defaults reference other values, we need to expand
        for key, value in config.items():
            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
                config[key] = value.format(**config)

        force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
        prefer_ipv6 = config.pop('prefer_ipv6', True)
        if not prefer_ipv6:
            force_ipv4 = True

        # Override global metrics config with more specific per-cloud
        # details.
        metrics_config = config.get('metrics', {})
        statsd_config = metrics_config.get('statsd', {})
        statsd_host = statsd_config.get('host') or self._statsd_host
        statsd_port = statsd_config.get('port') or self._statsd_port
        statsd_prefix = statsd_config.get('prefix') or self._statsd_prefix
        influxdb_config = metrics_config.get('influxdb', {})
        if influxdb_config:
            merged_influxdb = copy.deepcopy(self._influxdb_config)
            merged_influxdb.update(influxdb_config)
            influxdb_config = merged_influxdb
        else:
            influxdb_config = self._influxdb_config

        if cloud is None:
            cloud_name = ''
        else:
            cloud_name = str(cloud)
        return self._cloud_region_class(
            name=cloud_name,
            region_name=config['region_name'],
            config=config,
            extra_config=self.extra_config,
            force_ipv4=force_ipv4,
            auth_plugin=auth_plugin,
            openstack_config=self,
            session_constructor=self._session_constructor,
            app_name=self._app_name,
            app_version=self._app_version,
            cache_auth=self._cache_auth,
            cache_expiration_time=self._cache_expiration_time,
            cache_expirations=self._cache_expirations,
            cache_path=self._cache_path,
            cache_class=self._cache_class,
            cache_arguments=self._cache_arguments,
            password_callback=self._pw_callback,
            statsd_host=statsd_host,
            statsd_port=statsd_port,
            statsd_prefix=statsd_prefix,
            influxdb_config=influxdb_config,
        )

    def get_one_cloud(
        self,
        cloud: str | None = None,
        validate: bool = True,
        argparse: argparse_mod.Namespace | None = None,
        **kwargs: Any,
    ) -> cloud_region.CloudRegion:
        warnings.warn(
            "The 'get_one_cloud' method is a deprecated alias for 'get_one' "
            "and will be removed in a future release.",
            os_warnings.RemovedInSDK60Warning,
        )
        return self.get_one(
            cloud=cloud,
            validate=validate,
            argparse=argparse,
            **kwargs,
        )

    def get_one_cloud_osc(
        self,
        cloud: str | None = None,
        validate: bool = True,
        argparse: argparse_mod.Namespace | None = None,
        **kwargs: Any,
    ) -> cloud_region.CloudRegion:
        """Retrieve a single CloudRegion and merge additional options

        :param string cloud:
            The name of the configuration to load from clouds.yaml
        :param boolean validate:
            Validate the config. Setting this to False causes no auth plugin
            to be created. It's really only useful for testing.
        :param Namespace argparse:
            An argparse Namespace object; allows direct passing in of
            argparse options to be added to the cloud config.  Values
            of None and '' will be removed.
        :param region_name: Name of the region of the cloud.
        :param kwargs: Additional configuration options

        :raises: keystoneauth1.exceptions.MissingRequiredOptions
            on missing required auth parameters
        """

        args = self._fix_args(kwargs, argparse=argparse)

        if cloud is None:
            if 'cloud' in args:
                cloud = args['cloud']
            else:
                cloud = self.default_cloud

        config = self._get_base_cloud_config(cloud)

        # Get region specific settings
        if 'region_name' not in args:
            args['region_name'] = ''
        region = self._get_region(cloud=cloud, region_name=args['region_name'])
        args['region_name'] = region['name']
        region_args = copy.deepcopy(region['values'])

        # Regions is a list that we can use to create a list of cloud/region
        # objects. It does not belong in the single-cloud dict
        config.pop('regions', None)

        # Can't just do update, because None values take over
        for arg_list in region_args, args:
            for key, val in iter(arg_list.items()):
                if val is not None:
                    if key == 'auth' and config[key] is not None:
                        config[key] = _auth_update(config[key], val)
                    else:
                        config[key] = val

        config = self.magic_fixes(config)

        # NOTE(dtroyer): OSC needs a hook into the auth args before the
        #                plugin is loaded in order to maintain backward-
        #                compatible behaviour
        config = self.auth_config_hook(config)

        if validate:
            loader = self._get_auth_loader(config)
            config = self._validate_auth_correctly(config, loader)
            auth_plugin = loader.load_from_options(**config['auth'])
        else:
            auth_plugin = None

        # If any of the defaults reference other values, we need to expand
        for key, value in config.items():
            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
                config[key] = value.format(**config)

        force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
        prefer_ipv6 = config.pop('prefer_ipv6', True)
        if not prefer_ipv6:
            force_ipv4 = True

        if cloud is None:
            cloud_name = ''
        else:
            cloud_name = str(cloud)
        return self._cloud_region_class(
            name=cloud_name,
            region_name=config['region_name'],
            config=config,
            extra_config=self.extra_config,
            force_ipv4=force_ipv4,
            auth_plugin=auth_plugin,
            openstack_config=self,
            cache_auth=self._cache_auth,
            cache_expiration_time=self._cache_expiration_time,
            cache_expirations=self._cache_expirations,
            cache_path=self._cache_path,
            cache_class=self._cache_class,
            cache_arguments=self._cache_arguments,
            password_callback=self._pw_callback,
        )

    @staticmethod
    def set_one_cloud(
        config_file: str,
        cloud: str,
        set_config: dict[str, Any] | None = None,
    ) -> None:
        """Set a single cloud configuration.

        :param string config_file:
            The path to the config file to edit. If this file does not exist
            it will be created.
        :param string cloud:
            The name of the configuration to save to clouds.yaml
        :param dict set_config: Configuration options to be set
        """

        set_config = set_config or {}
        cur_config = {}
        try:
            with open(config_file) as fh:
                cur_config = yaml.safe_load(fh)
        except OSError as e:
            # Not no such file
            if e.errno != 2:
                raise
            pass

        clouds_config = cur_config.get('clouds', {})
        cloud_config = _auth_update(clouds_config.get(cloud, {}), set_config)
        clouds_config[cloud] = cloud_config
        cur_config['clouds'] = clouds_config

        with open(config_file, 'w') as fh:
            yaml.safe_dump(cur_config, fh, default_flow_style=False)


if __name__ == '__main__':
    config = OpenStackConfig().get_all_clouds()
    for cloud in config:
        print_cloud = False
        if len(sys.argv) == 1:
            print_cloud = True
        elif len(sys.argv) == 3 and (
            sys.argv[1] == cloud.name and sys.argv[2] == cloud.region
        ):
            print_cloud = True
        elif len(sys.argv) == 2 and (sys.argv[1] == cloud.name):
            print_cloud = True

        if print_cloud:
            print(cloud.name, cloud.region, cloud.config)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ManageableSnapshot(resource.Resource):
    resource_key = 'manageable-snapshot'
    resources_key = 'manageable-snapshots'
    base_path = '/manageable_snapshots'

    # capabilities
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'host',
        'cluster',
        'marker',
        'limit',
        'offset',
        'sort',
    )

    # Properties
    #: The ID of this snapshot in Cinder, if already managed.
    cinder_id = resource.Body('cinder_id')
    #: Additional information about the snapshot.
    extra_info = resource.Body('extra_info')
    #: The reason the snapshot is not safe to manage, if applicable.
    reason_not_safe = resource.Body('reason_not_safe')
    #: A reference to identify the existing snapshot on the backend.
    reference = resource.Body('reference', type=dict)
    #: Whether it is safe to manage this snapshot.
    safe_to_manage = resource.Body('safe_to_manage', type=bool)
    #: The size of the snapshot in GiB.
    size = resource.Body('size', type=int)
    #: A reference to the source volume of the snapshot.
    source_reference = resource.Body('source_reference', type=dict)
